c/*@@
c  @file       Init.F
c  @date      
c  @author     Gabrielle Allen + Miguel Alcubierre
c  @desc
c     Report on BSSN parameters and construct all
c     BSSN variables from standard ADM variables.
c  @enddesc 
c@@*/

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

c/*@@
c  @routine    ADM_BSSN_Init
c  @date       
c  @author     Gabrielle Allen + Miguel Alcubierre
c  @desc 
c     Report on BSSN parameters and construct all
c     BSSN variables from standard ADM variables.
c  @enddesc
c@@*/

      subroutine ADM_BSSN_Init(CCTK_ARGUMENTS)

      USE ADM_BSSN_Scalars

      implicit none

      DECLARE_CCTK_PARAMETERS
      DECLARE_CCTK_ARGUMENTS
      DECLARE_CCTK_FUNCTIONS

      integer i,j,k
      integer nx,ny,nz
      integer ierr

      integer, dimension(3) :: sw

#include "CactusEinstein/ADMMacros/src/macro/ADM_Spacing_declare.h"

      CCTK_REAL :: pm4,fac,aux
      CCTK_REAL :: finf,npow
      CCTK_REAL :: zero,half,one,third,twelfth

      CCTK_INT  :: symtable
      CCTK_INT, dimension(1:6) :: symbnd
      logical,  dimension(1:6) :: doBC

      character(50) :: infoline

#include "CactusEinstein/ADMMacros/src/macro/TRK_declare.h"
#include "CactusEinstein/ADMMacros/src/macro/DETG_declare.h"
#include "macro/BSGAMMA_declare.h"


c     Define sw (stencil width)
c     -------------------------

      sw(1) = 1; sw(2) = 1; sw(3) = 1


c     Numbers
c     -------

      zero    = 0.0D0
      half    = 0.5D0
      one     = 1.0D0
      third   = 1.0D0/3.0D0
      twelfth = 1.0D0/12.0D0


c     Grid parameters
c     ---------------

      nx = cctk_lsh(1)
      ny = cctk_lsh(2)
      nz = cctk_lsh(3)

#include "CactusEinstein/ADMMacros/src/macro/ADM_Spacing.h"


c     Sanity check for stencil size
c     -----------------------------

      if (stencil_size.gt.0) then
         stencilx = stencil_size
         stencily = stencil_size
         stencilz = stencil_size
      else
         stencilx = stencil_sizex
         stencily = stencil_sizey
         stencilz = stencil_sizez
      end if

      if (cctk_nghostzones(1).lt.stencilx) then
         call CCTK_WARN(0,"Ghost zones on x direction are smaller than stencil size")
      end if

      if (cctk_nghostzones(2).lt.stencily) then
         call CCTK_WARN(0,"Ghost zones on y direction are smaller than stencil size")
      end if

      if (cctk_nghostzones(3).lt.stencilz) then
         call CCTK_WARN(0,"Ghost zones on z direction are smaller than stencil size")
      end if


c     If using second order upwind, check the stencil size
c     ----------------------------------------------------

      if (CCTK_EQUALS(advection,'upwind2')) then

         if ((stencilx.lt.2).or.(stencily.lt.2).or.(stencilz.lt.2)) then
            call CCTK_WARN(0,"To use advection=upwind2 you need stencil_size >= 2")
         end if

      end if

c     Check different BSSN options
c     ----------------------------

      call CCTK_INFO("Chosen options for BSSN:")

      if (forceKzero.eq.1) then

         evolveK = .FALSE.

         if (CCTK_EQUALS(lapse_evolution_method,'maximal')) then
            call CCTK_INFO("  K forced to zero")
         else if (CCTK_EQUALS(lapse_evolution_method,'coordgauge')) then
            call CCTK_INFO("  WARNING: K forced to zero with mixed slicing!")
         else
            call CCTK_WARN(0,"forceKzero=yes can only be used with maximal slicing")
         end if

      else if (freezeK.eq.1) then

         evolveK = .FALSE.

         call CCTK_INFO("  K is not evolved")

      else

         evolveK = .TRUE.

         if (CCTK_EQUALS(lapse_evolution_method,'maximal')) then
            call CCTK_INFO("With maximal slicing you should set forceKzero=yes")
         end if

      end if

      if (subtracttrA.eq.1) then
         call CCTK_INFO("  Trace of Aij subtracted each timestep")
      else
         call CCTK_INFO("  Trace of Aij allowed to drift")
      end if

      if (freezeGamma.eq.0) then
         call CCTK_INFO("  Gammas are evolved")
         evolveGamma = .TRUE.
      else
         call CCTK_INFO("  Gammas are not evolved")
         evolveGamma = .FALSE.
      end if

      if (RfromH.eq.1) then
         call CCTK_INFO("  Replacing R using Hamiltonian constraint")
      else
         call CCTK_INFO("  R is used in evolution")
      end if

      if (attenuate.eq.1) then
         call CCTK_INFO("  Using attenuated evolution equations.")
      end if 
      if (attenuate_gauge.eq.1) then
         call CCTK_INFO("  Using attenuated gauge evolution equations.")
      end if
      if (attenuate.ne.1.and.attenuate_gauge.ne.1)then
         call CCTK_INFO("  No attenuation.")
      end if

c     Check shift parameters
c     ----------------------

c     No shift.

      if (CCTK_EQUALS(shift_evolution_method,'none')) then

         call CCTK_INFO("  No shift terms")
         evolveShift = .FALSE.
c     Zero shift.

      else if (CCTK_EQUALS(shift_evolution_method,'zero')) then

         call CCTK_INFO("  Zero shift")
         evolveShift = .FALSE.
c     Static shift.

      else if (CCTK_EQUALS(shift_evolution_method,'static')) then

         call CCTK_INFO("  Static shift")
         evolveShift = .FALSE.

c     Time derivative of shift proportional to Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma0')) then

         call CCTK_INFO("  Gamma0 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma0 shift")
         end if

      else if (CCTK_EQUALS(shift_evolution_method,'gammaS')) then

         call CCTK_INFO("  Mighty GammaS shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gammaS shift")
         end if

c     Time derivative of shift proportional to time derivative of Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma1')) then

         call CCTK_INFO("  Gamma1 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma1 shift")
         end if

c     Second time derivative of shift proportional to time derivative of Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma2')) then

         call CCTK_INFO("  Gamma2 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma2 shift")
         end if

c     Hyperbolic version of minimal distortion.

      else if (CCTK_EQUALS(shift_evolution_method,'hyp_mindist')) then

         call CCTK_INFO("  Hyperbolic minimal distortion shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with hyp_mindist shift")
         end if

c     Area locking shift.  Since this is not implemented via an
c     evolution equation we set evolveShift to .false.

c      else if (CCTK_EQUALS(shift_evolution_method,'locking')) then
c
c         call CCTK_INFO("  Area locking shift")
c         evolveShift = .FALSE.
c
c     Area locking shift via evolution equation.

c      else if (CCTK_EQUALS(shift_evolution_method,'dt_locking')) then
c
c         call CCTK_INFO("  Area locking shift via evolution equation")
c         evolveShift = .TRUE.
c
c     Unknown shift.  Set evolveShift = .FALSE.

      else

         evolveShift = .FALSE.

      end if


c     Initialize BSSN variabes to flat space
c     --------------------------------------

      ADM_BS_phi = 0.0d0

      ADM_BS_gxx = 1.0d0
      ADM_BS_gyy = 1.0d0
      ADM_BS_gzz = 1.0d0
      ADM_BS_gxy = 0.0d0
      ADM_BS_gxz = 0.0d0
      ADM_BS_gyz = 0.0d0

      ADM_BS_K   = 0.0d0

      ADM_BS_Axx = 0.0d0
      ADM_BS_Ayy = 0.0d0
      ADM_BS_Azz = 0.0d0
      ADM_BS_Axy = 0.0d0
      ADM_BS_Axz = 0.0d0
      ADM_BS_Ayz = 0.0d0

      ADM_BS_Gx  = 0.0d0
      ADM_BS_Gy  = 0.0d0
      ADM_BS_Gz  = 0.0d0


c     Calculate {phi,gij,K,Aij}
c     -------------------------

      do k=1,nz
         do j=1,ny
            do i=1,nx

c              Calculate BSSN phi from the cactus conformal metric (Eq.11)
c              -----------------------------------------------------------

#include "CactusEinstein/ADMMacros/src/macro/DETG_guts.h"

               ADM_BS_phi(i,j,k) = twelfth*log(DETG_DETCG) 


c              Calculate BSSN gij (Eq.10)
c              --------------------------

               pm4 = 1.0D0/DETG_DETCG**third

               ADM_BS_gxx(i,j,k) = pm4*gxx(i,j,k)
               ADM_BS_gxy(i,j,k) = pm4*gxy(i,j,k)
               ADM_BS_gxz(i,j,k) = pm4*gxz(i,j,k)
               ADM_BS_gyy(i,j,k) = pm4*gyy(i,j,k)
               ADM_BS_gyz(i,j,k) = pm4*gyz(i,j,k)
               ADM_BS_gzz(i,j,k) = pm4*gzz(i,j,k)


c              Calculate BSSN Aij from physical Kij (Eq.13)
c              --------------------------------------------

#include "CactusEinstein/ADMMacros/src/macro/TRK_guts.h"

               IF (conformal_state == 0) THEN
                  fac = pm4
               ELSE
                  fac = pm4/psi(i,j,k)**4
               ENDIF

               aux = third*TRK_TRK

               ADM_BS_Axx(i,j,k) = fac*kxx(i,j,k) - aux*ADM_BS_gxx(i,j,k)
               ADM_BS_Axy(i,j,k) = fac*kxy(i,j,k) - aux*ADM_BS_gxy(i,j,k)
               ADM_BS_Axz(i,j,k) = fac*kxz(i,j,k) - aux*ADM_BS_gxz(i,j,k)
               ADM_BS_Ayy(i,j,k) = fac*kyy(i,j,k) - aux*ADM_BS_gyy(i,j,k)
               ADM_BS_Ayz(i,j,k) = fac*kyz(i,j,k) - aux*ADM_BS_gyz(i,j,k)
               ADM_BS_Azz(i,j,k) = fac*kzz(i,j,k) - aux*ADM_BS_gzz(i,j,k)


c              Calculate BSSN variable K
c              -------------------------

               ADM_BS_K(i,j,k) = TRK_TRK

            end do
         end do
      end do

#include "CactusEinstein/ADMMacros/src/macro/DETG_undefine.h"
#include "CactusEinstein/ADMMacros/src/macro/TRK_undefine.h"


c     Calculate Gammas
c     ----------------

c     We need to calculate the Gammas in a separate loop since
c     since they require derivatives of the conformal metric. 

c     Interior.

      do k=2,nz-1
         do j=2,ny-1
            do i=2,nx-1

c     Now set the local_spatial_order depending on the position
c     relative to the boundaries or the excision region

               if (spatial_order > 2) then

                 if ( excise /= 0 .and. use_mask /= 0 ) then
                   if ( (i < 3).or.(i > cctk_lsh(1) - 2).or.
     .                  (j < 3).or.(j > cctk_lsh(2) - 2).or.
     .                  (k < 3).or.(k > cctk_lsh(3) - 2)) then
                     local_spatial_order = 2
                   else if (minval(emask(i-2:i+2,j-2:j+2,k-2:k+2)) < 0.75d0) then
                     local_spatial_order = 2
                   else
                     local_spatial_order = spatial_order
                   end if
                 else
                   if ( (i < 3).or.(i > cctk_lsh(1) - 2).or.
     .                  (j < 3).or.(j > cctk_lsh(2) - 2).or.
     .                  (k < 3).or.(k > cctk_lsh(3) - 2) ) then
                    local_spatial_order = 2
                   else
                     local_spatial_order = spatial_order
                   end if
                 end if
               end if

#include "macro/BSGAMMA_guts.h"

               ADM_BS_Gx(i,j,k) = BSGAMMA_GAMMAX
               ADM_BS_Gy(i,j,k) = BSGAMMA_GAMMAY
               ADM_BS_Gz(i,j,k) = BSGAMMA_GAMMAZ

            end do
         end do
      end do

#include "macro/BSGAMMA_undefine.h"

c     Apply boundary conditions to get Gammas on physical boundaries.
c     Notice that we only want to apply boundary conditions to the Gammas,
c     all other variables have been defined point-wise all the way to the
c     boundaries from the original ADM quantities (which SHOULD be
c     correctly defined all the way to the boundaries by the initial data
c     routines).  Here I use cubic extrapolation (though rational extrapolation
c     might be better).

      symtable = SymmetryTableHandleForGrid (cctkGH)
      if (symtable < 0) call CCTK_WARN (0, "internal error")
      call Util_TableGetIntArray (ierr, int(symtable), 6, symbnd, "symmetry_handle")
      if (ierr /= 6) call CCTK_WARN (0, "internal error")
      doBC = cctk_bbox /= 0 .and. symbnd < 0
      if (periodic.ne.0 .and. periodic_x.ne.0) then
         doBC(1) = .false.
         doBC(2) = .false.
      end if
      if (periodic.ne.0 .and. periodic_y.ne.0) then
         doBC(3) = .false.
         doBC(4) = .false.
      end if
      if (periodic.ne.0 .and. periodic_z.ne.0) then
         doBC(5) = .false.
         doBC(6) = .false.
      end if

      if (doBC(1)) then
         if (cctk_lsh(1) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the lower x direction")
         ADM_BS_Gx(1,:,:) = 4.0D0*ADM_BS_Gx(2,:,:) - 6.0D0*ADM_BS_Gx(3,:,:)
     .                    + 4.0D0*ADM_BS_Gx(4,:,:) - 1.0D0*ADM_BS_Gx(5,:,:)
         ADM_BS_Gy(1,:,:) = 4.0D0*ADM_BS_Gy(2,:,:) - 6.0D0*ADM_BS_Gy(3,:,:)
     .                    + 4.0D0*ADM_BS_Gy(4,:,:) - 1.0D0*ADM_BS_Gy(5,:,:)
         ADM_BS_Gz(1,:,:) = 4.0D0*ADM_BS_Gz(2,:,:) - 6.0D0*ADM_BS_Gz(3,:,:)
     .                    + 4.0D0*ADM_BS_Gz(4,:,:) - 1.0D0*ADM_BS_Gz(5,:,:)
      end if

      if (doBC(3)) then
         if (cctk_lsh(2) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the lower y direction")
         ADM_BS_Gx(:,1,:) = 4.0D0*ADM_BS_Gx(:,2,:) - 6.0D0*ADM_BS_Gx(:,3,:)
     .                    + 4.0D0*ADM_BS_Gx(:,4,:) - 1.0D0*ADM_BS_Gx(:,5,:)
         ADM_BS_Gy(:,1,:) = 4.0D0*ADM_BS_Gy(:,2,:) - 6.0D0*ADM_BS_Gy(:,3,:)
     .                    + 4.0D0*ADM_BS_Gy(:,4,:) - 1.0D0*ADM_BS_Gy(:,5,:)
         ADM_BS_Gz(:,1,:) = 4.0D0*ADM_BS_Gz(:,2,:) - 6.0D0*ADM_BS_Gz(:,3,:)
     .                    + 4.0D0*ADM_BS_Gz(:,4,:) - 1.0D0*ADM_BS_Gz(:,5,:)
      end if

      if (doBC(5)) then
         if (cctk_lsh(3) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the lower z direction")
         ADM_BS_Gx(:,:,1) = 4.0D0*ADM_BS_Gx(:,:,2) - 6.0D0*ADM_BS_Gx(:,:,3)
     .                    + 4.0D0*ADM_BS_Gx(:,:,4) - 1.0D0*ADM_BS_Gx(:,:,5)
         ADM_BS_Gy(:,:,1) = 4.0D0*ADM_BS_Gy(:,:,2) - 6.0D0*ADM_BS_Gy(:,:,3)
     .                    + 4.0D0*ADM_BS_Gy(:,:,4) - 1.0D0*ADM_BS_Gy(:,:,5)
         ADM_BS_Gz(:,:,1) = 4.0D0*ADM_BS_Gz(:,:,2) - 6.0D0*ADM_BS_Gz(:,:,3)
     .                    + 4.0D0*ADM_BS_Gz(:,:,4) - 1.0D0*ADM_BS_Gz(:,:,5)
      end if

      if (doBC(2)) then
         if (cctk_lsh(1) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the upper x direction")
         ADM_BS_Gx(nx,:,:) = 4.0D0*ADM_BS_Gx(nx-1,:,:) - 6.0D0*ADM_BS_Gx(nx-2,:,:)
     .                     + 4.0D0*ADM_BS_Gx(nx-3,:,:) - 1.0D0*ADM_BS_Gx(nx-4,:,:)
         ADM_BS_Gy(nx,:,:) = 4.0D0*ADM_BS_Gy(nx-1,:,:) - 6.0D0*ADM_BS_Gy(nx-2,:,:)
     .                     + 4.0D0*ADM_BS_Gy(nx-3,:,:) - 1.0D0*ADM_BS_Gy(nx-4,:,:)
         ADM_BS_Gz(nx,:,:) = 4.0D0*ADM_BS_Gz(nx-1,:,:) - 6.0D0*ADM_BS_Gz(nx-2,:,:)
     .                     + 4.0D0*ADM_BS_Gz(nx-3,:,:) - 1.0D0*ADM_BS_Gz(nx-4,:,:)
      end if

      if (doBC(4)) then
         if (cctk_lsh(2) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the upper y direction")
         ADM_BS_Gx(:,ny,:) = 4.0D0*ADM_BS_Gx(:,ny-1,:) - 6.0D0*ADM_BS_Gx(:,ny-2,:)
     .                     + 4.0D0*ADM_BS_Gx(:,ny-3,:) - 1.0D0*ADM_BS_Gx(:,ny-4,:)
         ADM_BS_Gy(:,ny,:) = 4.0D0*ADM_BS_Gy(:,ny-1,:) - 6.0D0*ADM_BS_Gy(:,ny-2,:)
     .                     + 4.0D0*ADM_BS_Gy(:,ny-3,:) - 1.0D0*ADM_BS_Gy(:,ny-4,:)
         ADM_BS_Gz(:,ny,:) = 4.0D0*ADM_BS_Gz(:,ny-1,:) - 6.0D0*ADM_BS_Gz(:,ny-2,:)
     .                     + 4.0D0*ADM_BS_Gz(:,ny-3,:) - 1.0D0*ADM_BS_Gz(:,ny-4,:)
      end if

      if (doBC(6)) then
         if (cctk_lsh(3) < 6) call CCTK_WARN (0, "local grid size is too small for the Gamma boundary conditions in the upper z direction")
         ADM_BS_Gx(:,:,nz) = 4.0D0*ADM_BS_Gx(:,:,nz-1) - 6.0D0*ADM_BS_Gx(:,:,nz-2)
     .                     + 4.0D0*ADM_BS_Gx(:,:,nz-3) - 1.0D0*ADM_BS_Gx(:,:,nz-4)
         ADM_BS_Gy(:,:,nz) = 4.0D0*ADM_BS_Gy(:,:,nz-1) - 6.0D0*ADM_BS_Gy(:,:,nz-2)
     .                     + 4.0D0*ADM_BS_Gy(:,:,nz-3) - 1.0D0*ADM_BS_Gy(:,:,nz-4)
         ADM_BS_Gz(:,:,nz) = 4.0D0*ADM_BS_Gz(:,:,nz-1) - 6.0D0*ADM_BS_Gz(:,:,nz-2)
     .                     + 4.0D0*ADM_BS_Gz(:,:,nz-3) - 1.0D0*ADM_BS_Gz(:,:,nz-4)
      end if

c     Synchronize and apply symmetries.

c     Now apply sync in the scheduler.

c      call CCTK_SyncGroup(ierr, cctkGH,"adm_bssn::ADM_BSSN_gamma")

      call CartSymGN(ierr,cctkGH,"adm_bssn::ADM_BSSN_gamma")


c     Initialize time derivatives of lapse and shift
c     ----------------------------------------------

      ADM_BS_dtalp = 0.0D0

      if (evolveShift) then
         ADM_BS_Bx = 0.0D0
         ADM_BS_By = 0.0D0
         ADM_BS_Bz = 0.0D0
      end if


c     Initialize mask
c     ---------------

      if (excise.eq.1) then

         ADM_BS_mask = 1.0D0

         if ((CCTK_EQUALS(excisiontype,'lego')).and.(use_mask.ne.1)) then
            call CCTK_WARN(0,"To use lego excision you must have use_mask active!")
         end if

         ADM_BS_dirx = 0.0D0
         ADM_BS_diry = 0.0D0
         ADM_BS_dirz = 0.0D0

      end if

      
c     Save initial data if needed
c     ---------------------------

      if (CCTK_EQUALS(bound,'staticrad').or.(gaugetest.eq.1)) then

         ADM_BS_alp0   = alp
         ADM_BS_dtalp0 = ADM_BS_dtalp

         ADM_BS_betax0 = betax
         ADM_BS_betay0 = betay
         ADM_BS_betaz0 = betaz
         if (evolveShift) then
           ADM_BS_Bx0    = ADM_BS_Bx
           ADM_BS_By0    = ADM_BS_By
           ADM_BS_Bz0    = ADM_BS_Bz
         end if

         ADM_BS_phi0 = ADM_BS_phi
         ADM_BS_gxx0 = ADM_BS_gxx
         ADM_BS_gyy0 = ADM_BS_gyy
         ADM_BS_gzz0 = ADM_BS_gzz
         ADM_BS_gxy0 = ADM_BS_gxy
         ADM_BS_gxz0 = ADM_BS_gxz
         ADM_BS_gyz0 = ADM_BS_gyz

         ADM_BS_K0   = ADM_BS_K
         ADM_BS_Axx0 = ADM_BS_Axx
         ADM_BS_Ayy0 = ADM_BS_Ayy
         ADM_BS_Azz0 = ADM_BS_Azz
         ADM_BS_Axy0 = ADM_BS_Axy
         ADM_BS_Axz0 = ADM_BS_Axz
         ADM_BS_Ayz0 = ADM_BS_Ayz

c     TODO: ADM_BS_G?0 needs to have its boundary conditions re-applied
c     in the postrestrict bin
         ADM_BS_Gx0 = ADM_BS_Gx
         ADM_BS_Gy0 = ADM_BS_Gy
         ADM_BS_Gz0 = ADM_BS_Gz

      end if


c     End
c     ---

c     Initialize the sources to be zero
c     ---------------------------------

      adm_bs_salp = 0.d0
      adm_bs_sdtalp = 0.d0
      adm_bs_sbetax = 0.d0
      adm_bs_sbetay = 0.d0
      adm_bs_sbetaz = 0.d0
      adm_bs_sBx = 0.d0
      adm_bs_sBy = 0.d0
      adm_bs_sBz = 0.d0
      adm_bs_sphi = 0.d0
      adm_bs_sgxx = 0.d0
      adm_bs_sgxy = 0.d0
      adm_bs_sgxz = 0.d0
      adm_bs_sgyy = 0.d0
      adm_bs_sgyz = 0.d0
      adm_bs_sgzz = 0.d0
      adm_bs_sK = 0.d0
      adm_bs_sAxx = 0.d0
      adm_bs_sAxy = 0.d0
      adm_bs_sAxz = 0.d0
      adm_bs_sAyy = 0.d0
      adm_bs_sAyz = 0.d0
      adm_bs_sAzz = 0.d0
      adm_bs_sGx = 0.d0
      adm_bs_sGy = 0.d0
      adm_bs_sGz = 0.d0

c     Initialize the previous time level to the same as the current
c     (necessary e.g. for excision boundaries)
c     -------------------------------------------------------------

      if (evolveShift) then
        betax_p = betax
        betay_p = betay
        betaz_p = betaz
        ADM_BS_Bx_p = ADM_BS_Bx
        ADM_BS_By_p = ADM_BS_By
        ADM_BS_Bz_p = ADM_BS_Bz
      end if
      alp_p = alp
      ADM_BS_dtalp_p = ADM_BS_dtalp
      ADM_BS_phi_p = ADM_BS_phi 
      ADM_BS_gxx_p = ADM_BS_gxx 
      ADM_BS_gxy_p = ADM_BS_gxy 
      ADM_BS_gxz_p = ADM_BS_gxz 
      ADM_BS_gyy_p = ADM_BS_gyy 
      ADM_BS_gyz_p = ADM_BS_gyz 
      ADM_BS_gzz_p = ADM_BS_gzz 
      ADM_BS_K_p = ADM_BS_K 
      ADM_BS_Axx_p = ADM_BS_Axx 
      ADM_BS_Axy_p = ADM_BS_Axy 
      ADM_BS_Axz_p = ADM_BS_Axz 
      ADM_BS_Ayy_p = ADM_BS_Ayy 
      ADM_BS_Ayz_p = ADM_BS_Ayz 
      ADM_BS_Azz_p = ADM_BS_Azz 
      ADM_BS_Gx_p = ADM_BS_Gx
      ADM_BS_Gy_p = ADM_BS_Gy
      ADM_BS_Gz_p = ADM_BS_Gz

      end subroutine ADM_BSSN_Init

c/*@@
c  @routine    ADM_BSSN_SetScalars
c  @date       
c  @author     Gabrielle Allen + Miguel Alcubierre, Ian Hawke
c  @desc 
c     Set up of logical scalars. Moved out of Init so that 
c     just these are reset after recovery
c  @enddesc
c@@*/

      subroutine ADM_BSSN_SetScalars(CCTK_ARGUMENTS)

      USE ADM_BSSN_Scalars

      implicit none

      DECLARE_CCTK_PARAMETERS
      DECLARE_CCTK_ARGUMENTS
      DECLARE_CCTK_FUNCTIONS

c     Check different BSSN options
c     ----------------------------

      call CCTK_INFO("Chosen options for BSSN:")

      if (forceKzero.eq.1) then

         evolveK = .FALSE.

         if (CCTK_EQUALS(lapse_evolution_method,'maximal')) then
            call CCTK_INFO("  K forced to zero")
         else if (CCTK_EQUALS(lapse_evolution_method,'coordgauge')) then
            call CCTK_INFO("  WARNING: K forced to zero with mixed slicing!")
         else
            call CCTK_WARN(0,"forceKzero=yes can only be used with maximal slicing")
         end if

      else if (freezeK.eq.1) then

         evolveK = .FALSE.

         call CCTK_INFO("  K is not evolved")

      else

         evolveK = .TRUE.

         if (CCTK_EQUALS(lapse_evolution_method,'maximal')) then
            call CCTK_INFO("With maximal slicing you should set forceKzero=yes")
         end if

      end if

      if (subtracttrA.eq.1) then
         call CCTK_INFO("  Trace of Aij subtracted each timestep")
      else
         call CCTK_INFO("  Trace of Aij allowed to drift")
      end if

      if (freezeGamma.eq.0) then
         call CCTK_INFO("  Gammas are evolved")
         evolveGamma = .TRUE.
      else
         call CCTK_INFO("  Gammas are not evolved")
         evolveGamma = .FALSE.
      end if

      if (RfromH.eq.1) then
         call CCTK_INFO("  Replacing R using Hamiltonian constraint")
      else
         call CCTK_INFO("  R is used in evolution")
      end if

      if (attenuate.eq.1) then
         call CCTK_INFO("  Using attenuated evolution equations.")
      endif 
      if (attenuate_gauge.eq.1) then
         call CCTK_INFO("  Using attenuated gauge evolution equations.")
      endif
      if (attenuate.ne.1.and.attenuate_gauge.ne.1)then
         call CCTK_INFO("  No attenuation.")
      end if

c     Check shift parameters
c     ----------------------

c     No shift.

      if (CCTK_EQUALS(shift_evolution_method,'none')) then

         call CCTK_INFO("  No shift terms")
         evolveShift = .FALSE.

c     Zero shift.

      else if (CCTK_EQUALS(shift_evolution_method,'zero')) then

         call CCTK_INFO("  Zero shift")
         evolveShift = .FALSE.

c     Static shift.

      else if (CCTK_EQUALS(shift_evolution_method,'static')) then

         call CCTK_INFO("  Static shift")
         evolveShift = .FALSE.

c     Time derivative of shift proportional to Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma0')) then

         call CCTK_INFO("  Gamma0 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma0 shift")
         end if

c     Time derivative of shift proportional to time derivative of Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma1')) then

         call CCTK_INFO("  Gamma1 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma1 shift")
         end if

c     Second time derivative of shift proportional to time derivative of Gammas.

      else if (CCTK_EQUALS(shift_evolution_method,'gamma2')) then

         call CCTK_INFO("  Gamma2 shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma2 shift")
         end if

      else if (CCTK_EQUALS(shift_evolution_method,'gammaS')) then

         call CCTK_INFO("  Mighty GammaS shift")
         evolveShift = .TRUE.
         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with gamma2 shift")
         end if

c     Hyperbolic version of minimal distortion.

      else if (CCTK_EQUALS(shift_evolution_method,'hyp_mindist')) then

         call CCTK_INFO("  Hyperbolic minimal distortion shift")
         evolveShift = .TRUE.

         if (.not.evolveGamma.and.evolveShift) then
            call CCTK_WARN(0,"freezeGamma=yes is incompatible with hyp_mindist shift")
         end if

c     Area locking shift.  Since this is not implemented via an
c     evolution equation we set evolveShift to .false.

c      else if (CCTK_EQUALS(shift_evolution_method,'locking')) then
c
c         call CCTK_INFO("  Area locking shift")
c         evolveShift = .FALSE.

c     Area locking shift via evolution equation.

c      else if (CCTK_EQUALS(shift_evolution_method,'dt_locking')) then
c
c         call CCTK_INFO("  Area locking shift via evolution equation")
c         evolveShift = .TRUE.

c     Unknown shift.  Set evolveShift = .FALSE.


      else

         evolveShift = .FALSE.

      end if

c     Use Shinkai&Yoneda adjusted BSSN equations

      if ((hisaaki_adjustments.eq.1).or.(baumgarte_adjustment.eq.1)) then
        calcAdjustment = .TRUE.
      else
        calcAdjustment = .FALSE.
      end if

      end subroutine ADM_BSSN_SetScalars
