/*@@
  @file    Lapse.F
  @date
  @author  Gabrielle Allen + Miguel Alcubierre
  @desc

     Set up coefficient "gauge" for lapse evolution. We will evolve
     one of two equations depending on the value of the parameter
     "lapsesource":

     1) lapsource = "standard":

        d_t alpha  =  gauge*(alpha K - div(beta)) + dissipation terms

     2) lapsource = "modified":

        d_t alpha  =  alpha*gauge*d_t K + dissipation terms

     The purpose of this routine is to set up the value of the
     coefficient "gauge".

  @enddesc
  @version $Header: /numrelcvs/AEIThorns/BSSN_MoL/src/Lapse.F,v 1.4 2002/10/11 12:22:48 hawke Exp $
@@*/

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"

      subroutine ADM_BSSN_Lapse(CCTK_ARGUMENTS,gauge)

      implicit none

      DECLARE_CCTK_ARGUMENTS
      DECLARE_CCTK_PARAMETERS

      integer CCTK_Equals
      integer Einstein_GetSlicingHandle

      CCTK_REAL, DIMENSION(cctk_lsh(1),cctk_lsh(2),cctk_lsh(3)) :: gauge


c     GEODESIC SLICING
c     ----------------

      if (active_slicing_handle.eq.Einstein_GetSlicingHandle("geodesic")) then

         alp   = 1.0D0
         gauge = 0.0D0


c     STATIC SLICING
c     --------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("static")) then

         gauge = 0.0D0


c     BONA-MASSO TYPE HARMONIC SLICING
c     --------------------------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("harmonic")) then

         gauge = alp*harmonic_f


c     1+LOG SLICING
c     -------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("1+log")) then

c        In the "kleing" (NCSA) style.

         if (CCTK_EQUALS(slicingFlavour,'kleing')) then

            if (keep_alp0.eq.1) then
               gauge = 2.0D0*ADM_BS_alp0
            else
               gauge = 2.0D0
            end if

c        In the Cactus style.

         else

            gauge = harmonic_f

         end if



c     SHOCK AVOIDING
c     --------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("shockavoid")) then

         gauge = alp*(1.0D0 + LapseCoeff/alp**2)


c     ZERO ORDER SHOCK AVOIDING
c     -------------------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("shockavoid0")) then

         gauge = alp*LapseCoeff**2/((2.0D0-LapseCoeff)
     .         + 2.0D0*(LapseCoeff-1.0D0)*alp)


c     FIRST ORDER SHOCK AVOIDING
c     --------------------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("shockavoid1")) then

         gauge = alp*LapseCoeff**3/((4.0D0-3.0D0*LapseCoeff)
     .         + alp*(1.0D0-LapseCoeff)*((4.0D0-LapseCoeff)*alp-8.0D0))     


c     MAXIMAL SLICING
c     ---------------

      else if (active_slicing_handle.eq.Einstein_GetSlicingHandle("maximal")) then

         gauge = 0.0D0


c     SANITY CHECK
c     ------------

      else if (Einstein_GetSlicingHandle(lapse_evolution_method).lt.0) then
        call CCTK_WARN(1,"thorn ADM_BSSN: slicing not registered by ADM_BSSN")
      else
        call CCTK_WARN(1,"thorn ADM_BSSN: slicing registered but not coded!")
      end if


c     END
c     ---

      end subroutine ADM_BSSN_Lapse



c     ==================================================================



      subroutine ADM_BSSN_alp0_init(CCTK_ARGUMENTS)

      implicit none

      DECLARE_CCTK_ARGUMENTS
      DECLARE_CCTK_PARAMETERS

      if (keep_alp0.eq.1) then
         call CCTK_INFO('Saving the initial lapse in ADM_BS_alp0')
         ADM_BS_alp0 = alp
      endif

      end subroutine ADM_BSSN_alp0_init
