 /*@@
   @file      Whisky_SphericalVelocity.F90
   @date      Sat Jan 26 02:03:56 2002
   @author    Ian Hawke
   @desc 
   Calculate the velocity in spherical polar coordinates for 
   analysis output.
   @enddesc 
 @@*/

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"
#include "cctk_Functions.h"

 /*@@
   @routine    Whisky_SphericalVelocity
   @date       Sat Jan 26 02:04:21 2002
   @author     Ian Hawke
   @desc 
   Calculate the velocity in spherical polar coordinates for 
   analysis output.
   @enddesc 
   @calls     
   @calledby   
   @history 

   @endhistory 

@@*/

subroutine Whisky_SphericalVelocity(CCTK_ARGUMENTS)
      
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  integer i, j, k, nx, ny, nz
  
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)

  do k = 1, nz
    do j = 1, ny
      do i = 1, nx

        if (abs(r(i,j,k)) < 1.d-10) then
          velradial(i,j,k) = 0.d0
          veltheta(i,j,k) = 0.d0
          velphi(i,j,k) = 0.d0
        else
          velradial(i,j,k) = ( x(i,j,k) * velx(i,j,k) + &
               y(i,j,k) * vely(i,j,k) + &
               z(i,j,k) * velz(i,j,k) ) / r(i,j,k)
          if (abs(x(i,j,k) * y(i,j,k)) < 1.d-10) then
            veltheta(i,j,k) = 0.d0
          else
            veltheta(i,j,k) = ( -y(i,j,k) * velx(i,j,k) + &
                 x(i,j,k) * vely(i,j,k) ) / &
                 (x(i,j,k)**2 + y(i,j,k)**2)
          end if
!!$          I cannot be bothered to write this one out yet.
          velphi(i,j,k) = 0.d0
        end if
      
      end do
    end do
  end do

end subroutine Whisky_SphericalVelocity

 /*@@
   @routine    Whisky_AngularMomentum
   @date       Sat Jan 26 02:04:21 2002
   @author     Ian Hawke
   @desc 
   Calculate the angular momentum (a la Shibata, PRD60 104052 eqn 2.25)
   @enddesc 
   @calls     
   @calledby   
   @history 

   @endhistory 

@@*/

subroutine Whisky_AngularMomentum(CCTK_ARGUMENTS)
      
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  CCTK_INT :: i, j, k, nx, ny, nz
  CCTK_REAL :: uxx, uxy, uxz, uyy, uyz, uzz, det
  CCTK_REAL :: EightPi, SixteenPi, TwelvePi

  EightPi = 32.d0 * atan(1.d0)
  TwelvePi = 48.d0 * atan(1.d0)
  SixteenPi = 64.d0 * atan(1.d0)
  
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)

  do k = 2, nz-1
    do j = 2, ny-1
      do i = 2, nx-1

        call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),&
             gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)
        call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
             gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),gyy(i,j,k),&
             gyz(i,j,k),gzz(i,j,k)) 

      end do
    end do
  end do
  
end subroutine Whisky_AngularMomentum

 /*@@
   @routine    Whisky_Distortion
   @date       Sat Feb 22 15:08:16 2003
   @author     Ian Hawke
   @desc 
   Calculates the distortion scalar. See Shibata et al., astro-ph/0005378.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

 @@*/

subroutine Whisky_Distortion(CCTK_ARGUMENTS)
      
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  DECLARE_CCTK_FUNCTIONS
  
  CCTK_INT :: ierr, Reduction_Handle, VarIndex

  call CCTK_ReductionHandle(Reduction_Handle, "sum")
  call CCTK_VarIndex(VarIndex, "Whisky::dens")

  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Whisky_ConservedM0, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate M0")
  end if

  call CCTK_VarIndex(VarIndex, "Whisky::Whisky_temp")
  
  Whisky_temp = dens*x**2
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Whisky_xrms, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate the xrms array")
  end if

  Whisky_temp = dens*y**2
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Whisky_yrms, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate the yrms array")
  end if

  Whisky_temp = dens*z**2
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Whisky_zrms, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate the zrms array")
  end if

  Whisky_xrms = sqrt(Whisky_xrms / Whisky_ConservedM0)
  Whisky_yrms = sqrt(Whisky_yrms / Whisky_ConservedM0)
  Whisky_zrms = sqrt(Whisky_zrms / Whisky_ConservedM0)

  Whisky_DistortionParameter = 2.d0 * (Whisky_xrms - Whisky_yrms) / &
       &(Whisky_xrms + Whisky_yrms)

end subroutine Whisky_Distortion

 /*@@
   @routine    Whisky_QECheck
   @date       Sat Feb 22 20:51:32 2003
   @author     Ian Hawke
   @desc 
   Uses the measure from Miller and Suen (gr-qc/0301112) to
   see how close the stars are to quasiequilibrium.

   Q_{ab} = 2 ( D_{(a} u_{b)} + u_{(a}a_{b)} )

   where D is the 4-covariant derivative and

   a_a = u_b D^b u_a

   a^a = u^b D_b u^a

   Projecting Q_{ab} into spatial components Q_{ij} and taking the
   norm to be

   |Q_{ij}| = sqrt{ max eval (Q_{ij}Q^j_k) }

   the scalar measure Q is found by normalizing w.r.t the maximum of the
   norms of the principle parts

   Q_{1ab} = 2 D_{(a}u_{b)},     Q_{2ab} = 2 u_{(a}a_{b)}

   Finally a mass weighted integral is taken.

   
   D^b u_a = g^{bc} (d_c u_a - Gamma^d_{ca} u_d)

   D_b u^a = d_b u^a + Gamma^a_{bc} u^c

   u^0 = w_lorentz / alp

   u^i = w_lorentz * (vel^i - beta^i / alp)

   

   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

 @@*/

!!$subroutine Whisky_QECheck(CCTK_ARGUMENTS)
!!$      
!!$  implicit none
!!$  
!!$  DECLARE_CCTK_ARGUMENTS
!!$  DECLARE_CCTK_PARAMETERS
!!$  DECLARE_CCTK_FUNCTIONS
!!$  
!!$  CCTK_INT :: ierr, Reduction_Handle_Sum, Reduction_Handle_Norm1, VarIndex
!!$
!!$  CCTK_INT :: i, j, k, nx, ny, nz
!!$
!!$#include "CactusEinstein/ADMMacros/src/macro/CHR2_declare.h"
!!$
!!$  call CCTK_ReductionHandle(Reduction_Handle_Sum, "sum")
!!$  call CCTK_ReductionHandle(Reduction_Handle_Norm1, "norm1")
!!$
!!$  nx = cctk_lsh(1)
!!$  ny = cctk_lsh(2)
!!$  nz = cctk_lsh(3)
!!$
!!$  do k = 1, nz
!!$    do j = 1, ny
!!$      do i = 1, nx
!!$
!!$        velxlow = gxx(i,j,k)*velx(i,j,k) + gxy(i,j,k)*vely(i,j,k) &
!!$             + gxz(i,j,k)*velz(i,j,k)
!!$        velylow = gxy(i,j,k)*velx(i,j,k) + gyy(i,j,k)*vely(i,j,k) &
!!$             + gyz(i,j,k)*velz(i,j,k)
!!$        velzlow = gxz(i,j,k)*velx(i,j,k) + gyz(i,j,k)*vely(i,j,k) &
!!$             + gzz(i,j,k)*velz(i,j,k)
!!$
!!$        if (shift_state .ne. 0) then
!!$
!!$          betaxlow = gxx(i,j,k)*betax(i,j,k) + gxy(i,j,k)*betay(i,j,k) &
!!$               + gxz(i,j,k)*betaz(i,j,k)
!!$          betaylow = gxy(i,j,k)*betax(i,j,k) + gyy(i,j,k)*betay(i,j,k) &
!!$               + gyz(i,j,k)*betaz(i,j,k)
!!$          betazlow = gxz(i,j,k)*betax(i,j,k) + gyz(i,j,k)*betay(i,j,k) &
!!$               + gzz(i,j,k)*betaz(i,j,k)
!!$
!!$          beta2 = betax(i,j,k)*betaxlow + betay(i,j,k)*betaylow + &
!!$               betaz(i,j,k)*betazlow 
!!$
!!$        else
!!$
!!$          betaxlow = 0.0D0
!!$          betaylow = 0.0D0
!!$          betazlow = 0.0D0
!!$          
!!$          beta2 = 0.0D0
!!$          
!!$        end if
!!$
!!$        utlow(i,j,k) = (-alp(i,j,k) + velx(i,j,k)*betaxlow + &
!!$             vely(i,j,k)*betaylow + velz(i,j,k)*betazlow)
!!$
!!$        uxlow(i,j,k) = velxlow
!!$        uylow(i,j,k) = velylow
!!$        uzlow(i,j,k) = velzlow
!!$
!!$      end do
!!$    end do
!!$  end do
!!$
!!$
!!$  do k = 2, nz - 1
!!$    do j = 2, ny - 1
!!$      do i = 2, nx - 1
!!$
!!$#include "CactusEinstein/ADMMacros/src/macro/CHR2_guts.h"
!!$
!!$        cdxutlow = ihdx * (utlow(i+1,j,k) - utlow(i-1,j,k)) - &
!!$             CHR2_XXX * uxlow(i,j,k) - CHR2_XY * uylow(i,j,k) - &
!!$             CHR2_XXZ * uzlow(i,j,k) 
!!$  
!!$      end do
!!$    end do
!!$  end do
!!$
!!$#include "CactusEinstein/ADMMacros/src/macro/CHR2_undefine.h"
!!$
!!$end subroutine Whisky_QECheck

 /*@@
   @routine    Whisky_Quadrupole
   @date       Tue Mar 18 12:47:12 2003
   @author     Ian Hawke, Luca Baiotti
   @desc 
   Calculates the first and the second time derivatives of the trace-free quadrupole
   moment as given by equation 37 and 39 of Finn and Evans (ApJ 351:588 (1990))
   Note that there does not exist an unambiguous definition of the
   velocities in the GR case so we fudge it.

   The units of Quadrupole_I?? are M L^2 T^{-1}, so to convert from
   Cactus to cgs you multiply by 8.806e48

   The units of Quadrupole_II?? are M L^2 T^{-2}, so to convert from
   Cactus to cgs you multiply by 1.787e54

   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/

subroutine Whisky_Quadrupole(CCTK_ARGUMENTS)
      
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  CCTK_INT :: i, j, k, l
  CCTK_REAL :: vx, vy, vz, &
       dnewton_pot_dx, dnewton_pot_dy, dnewton_pot_dz, &
       dnewton_pot_dx_up, dnewton_pot_dy_up, dnewton_pot_dz_up, &
       fourPi, det, uxx, uyy, uzz, uxy, uxz, uyz
  CCTK_INT :: ierr, Reduction_Handle, VarIndex
  
  fourPi = 4.d0 * acos(-1.0)

 do k = 1, cctk_lsh(3)
    do j = 1, cctk_lsh(2)
      do i = 1, cctk_lsh(1)
        
        Integrand_xx(i,j,k) = rho(i,j,k) * (x(i,j,k)**2 - r(i,j,k)**2 / 3.d0)
        Integrand_yy(i,j,k) = rho(i,j,k) * (y(i,j,k)**2 - r(i,j,k)**2 / 3.d0)
        Integrand_zz(i,j,k) = rho(i,j,k) * (z(i,j,k)**2 - r(i,j,k)**2 / 3.d0)
        Integrand_xy(i,j,k) = rho(i,j,k) * x(i,j,k) * y(i,j,k)
        Integrand_xz(i,j,k) = rho(i,j,k) * x(i,j,k) * z(i,j,k)
        Integrand_yz(i,j,k) = rho(i,j,k) * y(i,j,k) * z(i,j,k)

!!$        Floor it to stop strange results with tiny numbers

        if (abs(Integrand_xx(i,j,k)) < 1.e-15) then
          Integrand_xx(i,j,k) = 0.d0
        end if
        if (abs(Integrand_yy(i,j,k)) < 1.e-15) then
          Integrand_yy(i,j,k) = 0.d0
        end if
        if (abs(Integrand_zz(i,j,k)) < 1.e-15) then
          Integrand_zz(i,j,k) = 0.d0
        end if
        if (abs(Integrand_xy(i,j,k)) < 1.e-15) then
          Integrand_xy(i,j,k) = 0.d0
        end if
        if (abs(Integrand_xz(i,j,k)) < 1.e-15) then
          Integrand_xz(i,j,k) = 0.d0
        end if
        if (abs(Integrand_yz(i,j,k)) < 1.e-15) then
          Integrand_yz(i,j,k) = 0.d0
        end if


      end do
    end do
  end do

  call CCTK_ReductionHandle(Reduction_Handle, "sum")


  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xx")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0xx, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0xx")
  end if

  Quadrupole_0xx = Quadrupole_0xx * &
       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yy")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0yy, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0yy")
  end if

  Quadrupole_0yy = Quadrupole_0yy * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_zz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0zz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0zz")
  end if

  Quadrupole_0zz = Quadrupole_0zz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xy")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0xy, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0xy")
  end if

  Quadrupole_0xy = Quadrupole_0xy * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0xz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0xz")
  end if

  Quadrupole_0xz = Quadrupole_0xz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_0yz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_0yz")
  end if

  Quadrupole_0yz = Quadrupole_0yz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)


!!$  !compute first time derivative of the trace-free quadrupole moment
!!$
!!$  do k = 1, cctk_lsh(3)
!!$    do j = 1, cctk_lsh(2)
!!$      do i = 1, cctk_lsh(1)
!!$
!!$        vx = velx(i,j,k) * (velx(i,j,k) * gxx(i,j,k) + &
!!$             vely(i,j,k) * gxy(i,j,k) + &
!!$             velz(i,j,k) * gxz(i,j,k))
!!$        vy = vely(i,j,k) * (velx(i,j,k) * gxy(i,j,k) + &
!!$             vely(i,j,k) * gyy(i,j,k) + &
!!$             velz(i,j,k) * gyz(i,j,k))
!!$        vz = velz(i,j,k) * (velx(i,j,k) * gxz(i,j,k) + &
!!$             vely(i,j,k) * gyz(i,j,k) + &
!!$             velz(i,j,k) * gzz(i,j,k))
!!$
!!$        Integrand_xx(i,j,k) = rho(i,j,k) * (2.d0 * x(i,j,k) * vx - &
!!$             y(i,j,k) * vy - z(i,j,k) * vz) / 3.d0
!!$        Integrand_yy(i,j,k) = rho(i,j,k) * (2.d0 * y(i,j,k) * vy - &
!!$             x(i,j,k) * vx - z(i,j,k) * vz) / 3.d0
!!$        Integrand_zz(i,j,k) = rho(i,j,k) * (2.d0 * z(i,j,k) * vz - &
!!$             x(i,j,k) * vx - y(i,j,k) * vy) / 3.d0
!!$        Integrand_xy(i,j,k) = &
!!$             rho(i,j,k) * (x(i,j,k) * vy + y(i,j,k) * vx) / 2.d0
!!$        Integrand_xz(i,j,k) = &
!!$             rho(i,j,k) * (x(i,j,k) * vz + z(i,j,k) * vx) / 2.d0
!!$        Integrand_yz(i,j,k) = &
!!$             rho(i,j,k) * (y(i,j,k) * vz + z(i,j,k) * vy) / 2.d0
!!$
!!$        !Floor it to stop strange results with tiny numbers
!!$
!!$        if (abs(Integrand_xx(i,j,k)) < 1.e-15) then
!!$          Integrand_xx(i,j,k) = 0.d0
!!$        end if
!!$        if (abs(Integrand_yy(i,j,k)) < 1.e-15) then
!!$          Integrand_yy(i,j,k) = 0.d0
!!$        end if
!!$        if (abs(Integrand_zz(i,j,k)) < 1.e-15) then
!!$          Integrand_zz(i,j,k) = 0.d0
!!$        end if
!!$        if (abs(Integrand_xy(i,j,k)) < 1.e-15) then
!!$          Integrand_xy(i,j,k) = 0.d0
!!$        end if
!!$        if (abs(Integrand_xz(i,j,k)) < 1.e-15) then
!!$          Integrand_xz(i,j,k) = 0.d0
!!$        end if
!!$        if (abs(Integrand_yz(i,j,k)) < 1.e-15) then
!!$          Integrand_yz(i,j,k) = 0.d0
!!$        end if
!!$
!!$      end do
!!$    end do
!!$  end do
!!$
!!$  call CCTK_ReductionHandle(Reduction_Handle, "sum")
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xx")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Ixx, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Ixx")
!!$  end if
!!$
!!$  Quadrupole_Ixx = Quadrupole_Ixx * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yy")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Iyy, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Iyy")
!!$  end if
!!$
!!$  Quadrupole_Iyy = Quadrupole_Iyy * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_zz")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Izz, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Izz")
!!$  end if
!!$
!!$  Quadrupole_Izz = Quadrupole_Izz * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xy")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Ixy, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Ixy")
!!$  end if
!!$
!!$  Quadrupole_Ixy = Quadrupole_Ixy * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xz")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Ixz, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Ixz")
!!$  end if
!!$
!!$  Quadrupole_Ixz = Quadrupole_Ixz * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)
!!$
!!$  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yz")
!!$  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
!!$       1, CCTK_VARIABLE_REAL, Quadrupole_Iyz, 1, VarIndex)
!!$  if (ierr .ne. 0) then
!!$    call CCTK_WARN(0, "Failed to integrate Quadrupole_Iyz")
!!$  end if
!!$
!!$  Quadrupole_Iyz = Quadrupole_Iyz * &
!!$       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  !compute second time derivative of the trace-free quadrupole moment

  Integrand_xx = 0.d0
  Integrand_yy = 0.d0
  Integrand_zz = 0.d0
  Integrand_xy = 0.d0
  Integrand_xz = 0.d0
  Integrand_yz = 0.d0

  do k = 1, cctk_lsh(3)
    do j = 1, cctk_lsh(2)
      do i = 1, cctk_lsh(1)

        !radial part of the metric
        local_grr(i,j,k) = (x(i,j,k)**2 * gxx(i,j,k) + &
             y(i,j,k)**2 * gyy(i,j,k) + &
             z(i,j,k)**2 * gzz(i,j,k) + &
             2.d0 * x(i,j,k) * y(i,j,k) * gxy(i,j,k) + &
             2.d0 * x(i,j,k) * z(i,j,k) * gxz(i,j,k) + &
             2.d0 * y(i,j,k) * z(i,j,k) * gyz(i,j,k)) / r(i,j,k)**2

        newton_pot(i,j,k) = 0.5d0 * (1.d0 - local_grr(i,j,k))

      end do
    end do
  end do

  do k = whisky_stencil, cctk_lsh(3) - whisky_stencil + 1
    do j = whisky_stencil, cctk_lsh(2) - whisky_stencil + 1
      do i = whisky_stencil, cctk_lsh(1) - whisky_stencil + 1

        call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),&
             gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)
        call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
             gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),gyy(i,j,k),&
             gyz(i,j,k),gzz(i,j,k)) 
           
        dnewton_pot_dx = (newton_pot(i+1,j,k) - newton_pot(i-1,j,k)) / &
             (2.d0 * CCTK_DELTA_SPACE(1))
        dnewton_pot_dy = (newton_pot(i,j+1,k) - newton_pot(i,j-1,k)) / &
             (2.d0 * CCTK_DELTA_SPACE(2))
        dnewton_pot_dz = (newton_pot(i,j,k+1) - newton_pot(i,j,k-1)) / &
             (2.d0 * CCTK_DELTA_SPACE(3))

        dnewton_pot_dx_up = dnewton_pot_dx * uxx + &
             dnewton_pot_dy * uxy + &
             dnewton_pot_dz * uxz
        dnewton_pot_dy_up = dnewton_pot_dx * uxy + &
             dnewton_pot_dy * uyy + &
             dnewton_pot_dz * uyz
        dnewton_pot_dz_up = dnewton_pot_dx * uxz + &
             dnewton_pot_dy * uyz + &
             dnewton_pot_dz * uzz

        vx = velx(i,j,k) * gxx(i,j,k) + &
             vely(i,j,k) * gxy(i,j,k) + &
             velz(i,j,k) * gxz(i,j,k)
        vy = velx(i,j,k) * gxy(i,j,k) + &
             vely(i,j,k) * gyy(i,j,k) + &
             velz(i,j,k) * gyz(i,j,k)
        vz = velx(i,j,k) * gxz(i,j,k) + &
             vely(i,j,k) * gyz(i,j,k) + &
             velz(i,j,k) * gzz(i,j,k)

        Integrand_xx(i,j,k) = rho(i,j,k) * (3.d0 * vx * vx - &
             vx * velx(i,j,k) - vy * vely(i,j,k) - vz * velz(i,j,k)) / 3.d0 + &
             (3.d0 * dnewton_pot_dx * dnewton_pot_dx - dnewton_pot_dx * dnewton_pot_dx_up - &
             dnewton_pot_dy * dnewton_pot_dy_up - &
             dnewton_pot_dz * dnewton_pot_dz_up) / (3.d0 * fourPi)
        Integrand_yy(i,j,k) = rho(i,j,k) * (3.d0 * vy * vy - &
             vy * vely(i,j,k) * vx * velx(i,j,k) - vz * velz(i,j,k)) / 3.d0 + &
             (3.d0 * dnewton_pot_dy * dnewton_pot_dy - dnewton_pot_dy * dnewton_pot_dy_up -&
             dnewton_pot_dx * dnewton_pot_dx_up - &
             dnewton_pot_dz * dnewton_pot_dz_up) / (3.d0 * fourPi)
        Integrand_zz(i,j,k) = rho(i,j,k) * (3.d0 * vz * vz - &
             vz * velz(i,j,k) * vy * vely(i,j,k) - vx * velx(i,j,k)) / 3.d0 + &
             (3.d0 * dnewton_pot_dz * dnewton_pot_dz - dnewton_pot_dz * dnewton_pot_dz_up -&
             dnewton_pot_dy * dnewton_pot_dy_up - &
             dnewton_pot_dx * dnewton_pot_dx_up) / (3.d0 * fourPi)
        Integrand_xy(i,j,k) = &
             rho(i,j,k) * (vx * vy) + (dnewton_pot_dx * dnewton_pot_dy) / fourPi
        Integrand_xz(i,j,k) = &
             rho(i,j,k) * (vx * vz) + (dnewton_pot_dx * dnewton_pot_dz) / fourPi
        Integrand_yz(i,j,k) = &
             rho(i,j,k) * (vy * vz) + (dnewton_pot_dy * dnewton_pot_dz) / fourPi

!!$        Floor it to stop strange results with tiny numbers

        if (abs(Integrand_xx(i,j,k)) < 1.e-15) then
          Integrand_xx(i,j,k) = 0.d0
        end if
        if (abs(Integrand_yy(i,j,k)) < 1.e-15) then
          Integrand_yy(i,j,k) = 0.d0
        end if
        if (abs(Integrand_zz(i,j,k)) < 1.e-15) then
          Integrand_zz(i,j,k) = 0.d0
        end if
        if (abs(Integrand_xy(i,j,k)) < 1.e-15) then
          Integrand_xy(i,j,k) = 0.d0
        end if
        if (abs(Integrand_xz(i,j,k)) < 1.e-15) then
          Integrand_xz(i,j,k) = 0.d0
        end if
        if (abs(Integrand_yz(i,j,k)) < 1.e-15) then
          Integrand_yz(i,j,k) = 0.d0
        end if
        
      end do
    end do
  end do 
  
  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xx")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIxx, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIxx")
  end if

  Quadrupole_IIxx = Quadrupole_IIxx * &
       CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yy")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIyy, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIyy")
  end if

  Quadrupole_IIyy = Quadrupole_IIyy * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_zz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIzz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIzz")
  end if

  Quadrupole_IIzz = Quadrupole_IIzz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xy")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIxy, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIxy")
  end if

  Quadrupole_IIxy = Quadrupole_IIxy * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_xz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIxz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIxz")
  end if

  Quadrupole_IIxz = Quadrupole_IIxz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

  call CCTK_VarIndex(VarIndex, "Whisky::Integrand_yz")
  call CCTK_Reduce(ierr, cctkGH, -1, Reduction_Handle, &
       1, CCTK_VARIABLE_REAL, Quadrupole_IIyz, 1, VarIndex)
  if (ierr .ne. 0) then
    call CCTK_WARN(0, "Failed to integrate Quadrupole_IIyz")
  end if

  Quadrupole_IIyz = Quadrupole_IIyz * &
        CCTK_DELTA_SPACE(1) * CCTK_DELTA_SPACE(2) * CCTK_DELTA_SPACE(3)

end subroutine Whisky_Quadrupole
