 /*@@
   @file      Whisky_RegisterVars.c
   @date      Sat Jan 26 01:06:01 2002
   @author    
   @desc 
   The routines for converting conservative to primitive variables.
   @enddesc 
 @@*/

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"

 /*@@
   @routine    Conservative2Primitive
   @date       Sat Jan 26 01:08:33 2002
   @author     Ian Hawke
   @desc 
   Wrapper routine that converts from conserved to primitive variables
   at every grid cell centre.
   @enddesc 
   @calls     
   @calledby   
   @history 
   Trimmed and altered from the GR3D routines, original author Mark Miller.
   @endhistory 

@@*/

subroutine Conservative2Primitive(CCTK_ARGUMENTS)
  
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  integer :: i, j, k, nx, ny, nz
  CCTK_REAL :: uxx, uxy, uxz, uyy, uyz, uzz, det, psi4pt
  logical :: epsnegative
  character(len=100) warnline
  
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
  
!!$  do k = whisky_stencil + 1, nz - whisky_stencil
!!$    do j = whisky_stencil + 1, ny - whisky_stencil
!!$      do i = whisky_stencil + 1, nx - whisky_stencil
  do k = 1, nz 
    do j = 1, ny 
      do i = 1, nx

        epsnegative = .false.

        if (conformal_state .eq. 0) then
          call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),&
               gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)
          call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
               gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),gyy(i,j,k),&
               gyz(i,j,k),gzz(i,j,k))        
        else
          psi4pt = psi(i,j,k)**4
          call SpatialDeterminant(psi4pt*gxx(i,j,k),psi4pt*gxy(i,j,k),&
               psi4pt*gxz(i,j,k),psi4pt*gyy(i,j,k),&
               psi4pt*gyz(i,j,k),psi4pt*gzz(i,j,k),det)
          call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
               psi4pt*gxx(i,j,k),psi4pt*gxy(i,j,k),psi4pt*gxz(i,j,k),&
               psi4pt*gyy(i,j,k),psi4pt*gyz(i,j,k),psi4pt*gzz(i,j,k))
        end if

        call Con2Prim_pt(whisky_eos_handle, dens(i,j,k),sx(i,j,k),sy(i,j,k), &
             sz(i,j,k),tau(i,j,k),rho(i,j,k),velx(i,j,k),vely(i,j,k), &
             velz(i,j,k),eps(i,j,k),press(i,j,k),w_lorentz(i,j,k), &
             uxx,uxy,uxz,uyy,uyz,uzz,det,x(i,j,k),y(i,j,k), &
             z(i,j,k),r(i,j,k),epsnegative,whisky_rho_min)

        if (epsnegative) then  
          call Con2Prim_ptPolytype(whisky_polytrope_handle, &
               dens(i,j,k), sx(i,j,k), sy(i,j,k), sz(i,j,k), &
               tau(i,j,k), rho(i,j,k), velx(i,j,k), vely(i,j,k), &
               velz(i,j,k), eps(i,j,k), press(i,j,k), w_lorentz(i,j,k), &
               uxx, uxy, uxz, uyy, uyz, uzz, det, x(i,j,k), y(i,j,k), &
               z(i,j,k), r(i,j,k),whisky_rho_min)  
        end if

        if(eps(i,j,k) .lt. 0.0d0) then
          call CCTK_WARN(1,'Con2Prim: stopping the code.')
          call CCTK_WARN(1, '   specific internal energy just went below 0! ')
          write(warnline,'(a20,3g16.7)') 'xyz location: ',&
               x(i,j,k),y(i,j,k),z(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,g16.7)') 'radius: ',r(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,3g16.7)') 'velocities: ',&
               velx(i,j,k),vely(i,j,k),velz(i,j,k)
          call CCTK_WARN(1,warnline)
          call CCTK_WARN(whisky_c2p_warnlevel, "Specific internal energy negative")
        endif
  
      end do
    end do
  end do
  
end subroutine Conservative2Primitive

/*@@
@routine    Con2Prim_pt
@date       Sat Jan 26 01:09:39 2002
@author     
@desc 
Given all the appropriate data, recover the primitive variables at
a single point.
@enddesc 
 @calls     
@calledby   
@history 

@endhistory 

@@*/

subroutine Con2Prim_pt(handle, dens, sx, sy, sz, tau, rho, velx, vely, &
     velz, epsilon, press, w_lorentz, uxx, uxy, uxz, uyy, &
     uyz, uzz, det, x, y, z, r, epsnegative,whisky_rho_min)
  
  implicit none
  
  DECLARE_CCTK_PARAMETERS
  
  CCTK_REAL dens, sx, sy, sz, tau, rho, velx, vely, velz, epsilon, &
       press, uxx, uxy, uxz, uyy, uyz, uzz, det, w_lorentz, x, &
       y, z, r, whisky_rho_min
  
  CCTK_REAL s2, c0, c1, c2, c3, c4, f, df, ftol, v2, w, vlowx, vlowy, &
       vlowz 
  CCTK_REAL dens_prim, tau_prim, sx_prim, sy_prim, sz_prim, test
  CCTK_INT count, i, handle
  CCTK_REAL udens, usx, usy, usz, utau, pold, pnew, epsold, epsnew, w2, &
       w2mhalf, temp1,drhobydpress,depsbydpress,dpressbydeps,dpressbydrho
  character(len=100) warnline
  logical epsnegative

#ifdef _EOS_BASE_INC_
#undef _EOS_BASE_INC_
#endif
#include "EOS_Base.inc"

!!$I''ve removed the setting to zero for the T3E that was here.
!!$It was a hack to start with, and you should get the same results
!!$in a cleaner fashion using the parameter 
!!$pugh::initialize_memory="zero"
  
!!$  Undensitize the variables 

  udens = dens /sqrt(det)
  usx = sx /sqrt(det)
  usy = sy /sqrt(det)
  usz = sz /sqrt(det)
  utau = tau /sqrt(det)
  s2 = usx*usx*uxx + usy*usy*uyy + usz*usz*uzz + 2.*usx*usy*uxy + &
       2.*usx*usz*uxz + 2.*usy*usz*uyz

!!$  Set initial guess for pressure:

  pold = max(1.d-10,EOS_Pressure(handle, rho, epsilon))

!!$  Check that the variables have a chance of being physical

  if( (utau + pold + udens)**2 - s2 .le. 0.0d0) then
    call CCTK_WARN(1,'Con2Prim_pt: physical variables unphysical!')
    write(warnline,'(a25,g15.6)') 'undensitized energy: ',utau
    call CCTK_WARN(1,warnline)
    write(warnline,'(a25,g15.6)') 'undensitized density: ',udens
    call CCTK_WARN(1,warnline)
    write(warnline,'(a25,g15.6)') 'undensitized s2: ',s2
    call CCTK_WARN(1,warnline)
    write(warnline,'(a25,g15.6)') 'pressure guess: ',pold
    call CCTK_WARN(1,warnline)
    write(warnline,'(a25,4g15.6)') 'coordinates: x,y,z,r:',x,y,z,r
    call CCTK_WARN(1,warnline)
    call CCTK_WARN(whisky_c2p_warnlevel, "Unphysical variables")
  endif
  
!!$  Calculate rho and epsilon 

  rho = udens * sqrt( (utau + pold + udens)**2 - s2)/(utau + pold + udens)
  w_lorentz = (utau + pold + udens) / sqrt( (utau + pold + udens)**2 - s2)
  epsilon = (sqrt( (utau + pold + udens)**2 - s2) - pold * w_lorentz - &
       udens)/udens
  
!!$  Calculate the function

  f = pold - EOS_Pressure(handle, rho, epsilon)

!!$Find the root
  
  count = 0
  pnew = pold
  do while ( ((abs(pnew - pold)/abs(pnew) .gt. whisky_perc_ptol) .and. &
       (abs(pnew - pold) .gt. whisky_del_ptol))  .or. &
       (count .lt. whisky_countmin))
    count = count + 1
    if (count .gt. whisky_countmax) then
      call CCTK_WARN(1, 'Con2Prim_pt(press): error: did not converge in ')
      write(warnline,'(a20,g12.7,a10)') '              ',&
           whisky_countmax,' steps'
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'de is ',dens
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sx is ',sx
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sy is ',sy
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sz is ',sz
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'en is ',tau
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'press func: ',f
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'deriv of press func: ',df
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'pnew: ',pnew
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'pold: ',pold
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'rho: ',rho
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'epsilon: ',epsilon
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'w_lorentz: ',w_lorentz
      call CCTK_WARN(1,warnline)

      write(warnline,'(a20,3g12.7)') 'xyz location: ',x,y,z
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'radius: ',r
      call CCTK_WARN(1,warnline)
      call CCTK_WARN(whisky_c2p_warnlevel, "Did not converge")
    endif

    dpressbydrho = EOS_DPressByDRho(handle,rho,epsilon)
    dpressbydeps = EOS_DPressByDEps(handle,rho,epsilon)
    temp1 = (utau+udens+pnew)**2 - s2
    drhobydpress = udens * s2 / (sqrt(temp1)*(udens+utau+pnew)**2)
    depsbydpress = pnew * s2 / (rho * (udens + utau + pnew) * temp1)
    df = 1.0d0 - dpressbydrho*drhobydpress - &
         dpressbydeps*depsbydpress

    pold = pnew
    pnew = pold - f/df
    
!!$    Recalculate primitive variables and function
       
    rho = udens * sqrt( (utau + pnew + udens)**2 - s2)/(utau + pnew + udens)
    w_lorentz = (utau + pnew + udens) / sqrt( (utau + pnew + udens)**2 - &
         s2)
    epsilon = (sqrt( (utau + pnew + udens)**2 - s2) - pnew * w_lorentz - &
         udens)/udens

    f = pnew - EOS_Pressure(handle, rho, epsilon)

  enddo
  
!!$  Polish the root?

  do i=1,whisky_polish

    dpressbydrho = EOS_DPressByDRho(handle,rho,epsilon)
    dpressbydeps = EOS_DPressByDEps(handle,rho,epsilon)
    temp1 = (utau+udens+pnew)**2 - s2
    drhobydpress = udens * s2 / (sqrt(temp1)*(udens+utau+pnew)**2)
    depsbydpress = pnew * s2 / (rho * (udens + utau + pnew) * temp1)
    df = 1.0d0 - dpressbydrho*drhobydpress - &
         dpressbydeps*depsbydpress
    pold = pnew
    pnew = pold - f/df
    
!!$    Recalculate primitive variables and function

    rho = udens * sqrt( (utau + pnew + udens)**2 - s2)/(utau + pnew + udens)
    w_lorentz = (utau + pnew + udens) / sqrt( (utau + pnew + udens)**2 - &
         s2)
    epsilon = (sqrt( (utau + pnew + udens)**2 - s2) - pnew * w_lorentz - &
         udens)/udens

    f = pold - EOS_Pressure(handle, rho, epsilon)

  enddo
      
!!$  Calculate primitive variables from root

  if (rho .le. whisky_rho_min) then
    rho = whisky_rho_min
    udens = rho
    dens = sqrt(det) * rho
    epsilon = (sqrt( (utau + pnew + udens)**2) - pnew - &
         udens)/udens
    sx = 0.d0
    sy = 0.d0
    sz = 0.d0
    s2 = 0.d0
    usx = 0.d0
    usy = 0.d0
    usz = 0.d0
    w_lorentz = 1.d0
  end if

  press = pnew
  vlowx = usx / ( (rho + rho*epsilon + press) * w_lorentz**2)
  vlowy = usy / ( (rho + rho*epsilon + press) * w_lorentz**2)
  vlowz = usz / ( (rho + rho*epsilon + press) * w_lorentz**2)
  velx = uxx * vlowx + uxy * vlowy + uxz * vlowz
  vely = uxy * vlowx + uyy * vlowy + uyz * vlowz
  velz = uxz * vlowx + uyz * vlowy + uzz * vlowz
  
!!$  Check and make sure we actually solved the equations.   This
!!$  can someday be commented out.

!!$  Note this only applies to ideal gas!

  dens_prim = dens - sqrt(det) * rho * w_lorentz
  tau_prim = tau - sqrt(det) * ((rho * (1.0d0 + epsilon) + press) * &
       w_lorentz**2  - press - rho * w_lorentz)
  sx_prim = sx-sqrt(det)*(rho*(1.0d0+epsilon) + press)*vlowx * &
       w_lorentz**2
  sy_prim = sy-sqrt(det)*(rho*(1.0d0+epsilon) + press)*vlowy * &
       w_lorentz**2
  sz_prim = sz-sqrt(det)*(rho*(1.0d0+epsilon) + press)*vlowz * &
       w_lorentz**2
  
  test = max(abs(dens_prim),abs(tau_prim))
  test = max(test,abs(sx_prim))
  test = max(test,abs(sy_prim))
  test = max(test,abs(sz_prim))
  
#ifndef SINGLE_PRECISION
  if ((test .ge. 1.0d-20).and.(test/rho .ge. 1.0d-10)) then
    call CCTK_WARN(1,'Con2Prim_pt: 1: error: ')
    write(warnline,'(a20,5g16.7)') 'Cons vars',dens_prim,tau_prim,&
         sx_prim,sy_prim,sz_prim
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'en : ',tau
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'det: ',det
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'rho: ',rho
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'epsilon: ',epsilon
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'press: ',press
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'w_lorentz: ',w_lorentz
    call CCTK_WARN(1,warnline)
    call CCTK_WARN(whisky_c2p_warnlevel, "Error stage 1")
  endif
#endif

!!$If all else fails, use the polytropic EoS

  if(epsilon .lt. 0.0d0) then
    epsnegative = .true.
  endif

end subroutine Con2Prim_pt

 /*@@
   @routine    Conservative2PrimitiveBoundaries
   @date       Tue Mar 12 18:04:40 2002
   @author     
   @desc 
   
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/


subroutine Conservative2PrimitiveBounds(CCTK_ARGUMENTS)
  
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  integer :: i, j, k, nx, ny, nz
  CCTK_REAL :: uxxl, uxyl, uxzl, uyyl, uyzl, uzzl,&
       uxxr, uxyr, uxzr, uyyr, uyzr, uzzr
  CCTK_REAL :: gxxl,gxyl,gxzl,gyyl,gyzl,gzzl,avg_detl,&
       gxxr,gxyr,gxzr,gyyr,gyzr,gzzr,avg_detr,psi4r,psi4l
  logical :: epsnegative
  character(len=100) warnline
 
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
  
  
  do k = whisky_stencil, nz - whisky_stencil + 1
    do j = whisky_stencil, ny - whisky_stencil + 1
      do i = whisky_stencil, nx - whisky_stencil + 1

        gxxl = 0.5d0 * (gxx(i,j,k) + gxx(i-xoffset,j-yoffset,k-zoffset))
        gxyl = 0.5d0 * (gxy(i,j,k) + gxy(i-xoffset,j-yoffset,k-zoffset))
        gxzl = 0.5d0 * (gxz(i,j,k) + gxz(i-xoffset,j-yoffset,k-zoffset))
        gyyl = 0.5d0 * (gyy(i,j,k) + gyy(i-xoffset,j-yoffset,k-zoffset))
        gyzl = 0.5d0 * (gyz(i,j,k) + gyz(i-xoffset,j-yoffset,k-zoffset))
        gzzl = 0.5d0 * (gzz(i,j,k) + gzz(i-xoffset,j-yoffset,k-zoffset))
        gxxr = 0.5d0 * (gxx(i,j,k) + gxx(i+xoffset,j+yoffset,k+zoffset))
        gxyr = 0.5d0 * (gxy(i,j,k) + gxy(i+xoffset,j+yoffset,k+zoffset))
        gxzr = 0.5d0 * (gxz(i,j,k) + gxz(i+xoffset,j+yoffset,k+zoffset))
        gyyr = 0.5d0 * (gyy(i,j,k) + gyy(i+xoffset,j+yoffset,k+zoffset))
        gyzr = 0.5d0 * (gyz(i,j,k) + gyz(i+xoffset,j+yoffset,k+zoffset))
        gzzr = 0.5d0 * (gzz(i,j,k) + gzz(i+xoffset,j+yoffset,k+zoffset))

        epsnegative = .false.

        if (conformal_state .eq. 0) then
          call SpatialDeterminant(gxxl,gxyl,gxzl,gyyl, gyzl,gzzl,avg_detl)
          call SpatialDeterminant(gxxr,gxyr,gxzr,gyyr, gyzr,gzzr,avg_detr)
          call UpperMetric(uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
               gxxl,gxyl,gxzl,gyyl,gyzl,gzzl)        
          call UpperMetric(uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
               gxxr,gxyr,gxzr,gyyr,gyzr,gzzr)        
        else
          psi4r = (0.5d0*(psi(i,j,k)+psi(i+xoffset,j+yoffset,k+zoffset)))**4
          psi4l = (0.5d0*(psi(i,j,k)+psi(i-xoffset,j-yoffset,k-zoffset)))**4

          call SpatialDeterminant(psi4l*gxxl,psi4l*gxyl,psi4l*gxzl,&
               psi4l*gyyl,psi4l*gyzl,psi4l*gzzl,avg_detl)
          call SpatialDeterminant(psi4r*gxxr,psi4r*gxyr,psi4r*gxzr,&
               psi4r*gyyr,psi4r*gyzr,psi4r*gzzr,avg_detr)
          call UpperMetric(uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
               psi4l*gxxl,psi4l*gxyl,psi4l*gxzl,psi4l*gyyl,psi4l*gyzl,&
               psi4l*gzzl)        
          call UpperMetric(uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
               psi4r*gxxr,psi4r*gxyr,psi4r*gxzr,psi4r*gyyr,psi4r*gyzr,&
               psi4r*gzzr)        
        end if

        call Con2Prim_pt(whisky_eos_handle, densminus(i,j,k),&
             sxminus(i,j,k),syminus(i,j,k),szminus(i,j,k),&
             tauminus(i,j,k),rhominus(i,j,k),velxminus(i,j,k),&
             velyminus(i,j,k),velzminus(i,j,k),epsminus(i,j,k),&
             pressminus(i,j,k),w_lorentzminus(i,j,k),&
             uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
             x(i,j,k)-0.5d0*CCTK_DELTA_SPACE(1),&
             y(i,j,k)-0.5d0*CCTK_DELTA_SPACE(2), &
             z(i,j,k)-0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),&
             epsnegative,whisky_rho_min)
        if (epsnegative) then
          call Con2Prim_ptPolytype(whisky_polytrope_handle, densminus(i,j,k),&
               sxminus(i,j,k),syminus(i,j,k),szminus(i,j,k),&
               tauminus(i,j,k),rhominus(i,j,k),velxminus(i,j,k),&
               velyminus(i,j,k),velzminus(i,j,k),epsminus(i,j,k),&
               pressminus(i,j,k),w_lorentzminus(i,j,k),&
               uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
               x(i,j,k)-0.5d0*CCTK_DELTA_SPACE(1),&
               y(i,j,k)-0.5d0*CCTK_DELTA_SPACE(2), &
               z(i,j,k)-0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),whisky_rho_min)
        end if

        if(epsminus(i,j,k) .lt. 0.0d0) then
          call CCTK_WARN(1,'Con2Prim: stopping the code.')
          call CCTK_WARN(1, '   specific internal energy just went below 0! ')
          write(warnline,'(a20,3g16.7)') 'xyz location: ',&
               x(i,j,k),y(i,j,k),z(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,g16.7)') 'radius: ',r(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,3g16.7)') 'velocities: ',&
               velxminus(i,j,k),velyminus(i,j,k),velzminus(i,j,k)
          call CCTK_WARN(1,warnline)
          call CCTK_WARN(0, "Specific internal energy negative")
        endif

        epsnegative = .false.
        call Con2Prim_pt(whisky_eos_handle, densplus(i,j,k),&
             sxplus(i,j,k),syplus(i,j,k),szplus(i,j,k),&
             tauplus(i,j,k),rhoplus(i,j,k),velxplus(i,j,k),&
             velyplus(i,j,k),velzplus(i,j,k),epsplus(i,j,k),&
             pressplus(i,j,k),w_lorentzplus(i,j,k),&
             uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
             x(i,j,k)+0.5d0*CCTK_DELTA_SPACE(1),&
             y(i,j,k)+0.5d0*CCTK_DELTA_SPACE(2), &
             z(i,j,k)+0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),&
             epsnegative,whisky_rho_min)
        if (epsnegative) then
          call Con2Prim_ptPolytype(whisky_polytrope_handle, densplus(i,j,k),&
               sxplus(i,j,k),syplus(i,j,k),szplus(i,j,k),&
               tauplus(i,j,k),rhoplus(i,j,k),velxplus(i,j,k),&
               velyplus(i,j,k),velzplus(i,j,k),epsplus(i,j,k),&
               pressplus(i,j,k),w_lorentzplus(i,j,k),&
               uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
               x(i,j,k)+0.5d0*CCTK_DELTA_SPACE(1),&
               y(i,j,k)+0.5d0*CCTK_DELTA_SPACE(2), &
               z(i,j,k)+0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),whisky_rho_min)
        end if

        if(epsplus(i,j,k) .lt. 0.0d0) then
          call CCTK_WARN(1,'Con2Prim: stopping the code.')
          call CCTK_WARN(1, '   specific internal energy just went below 0! ')
          write(warnline,'(a20,3g16.7)') 'xyz location: ',&
               x(i,j,k),y(i,j,k),z(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,g16.7)') 'radius: ',r(i,j,k)
          call CCTK_WARN(1,warnline)
          write(warnline,'(a20,3g16.7)') 'velocities: ',&
               velxplus(i,j,k),velyplus(i,j,k),velzplus(i,j,k)
          call CCTK_WARN(1,warnline)
          call CCTK_WARN(0, "Specific internal energy negative")
        endif
        
      end do
    end do
  end do
  
end subroutine Conservative2PrimitiveBounds

 /*@@
   @routine    Con2PrimPolytype
   @date       Tue Mar 19 11:43:06 2002
   @author     Ian Hawke
   @desc 
   All routines below are identical to those above, just
   specialised from polytropic type EOS.
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/


subroutine Conservative2PrimitivePolytype(CCTK_ARGUMENTS)
  
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  integer :: i, j, k, nx, ny, nz
  CCTK_REAL :: uxx, uxy, uxz, uyy, uyz, uzz, det, psi4pt
  
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
  
!!$  do k = whisky_stencil + 1, nz - whisky_stencil
!!$    do j = whisky_stencil + 1, ny - whisky_stencil
!!$      do i = whisky_stencil + 1, nx - whisky_stencil
  do k = 1, nz
    do j = 1, ny
      do i = 1, nx

        if (conformal_state .eq. 0) then
          call SpatialDeterminant(gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),&
               gyy(i,j,k),gyz(i,j,k),gzz(i,j,k),det)
          call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
               gxx(i,j,k),gxy(i,j,k),gxz(i,j,k),gyy(i,j,k),&
               gyz(i,j,k),gzz(i,j,k))        
        else
          psi4pt = psi(i,j,k)**4
          call SpatialDeterminant(psi4pt*gxx(i,j,k),psi4pt*gxy(i,j,k),&
               psi4pt*gxz(i,j,k),psi4pt*gyy(i,j,k),&
               psi4pt*gyz(i,j,k),psi4pt*gzz(i,j,k),det)
          call UpperMetric(uxx,uxy,uxz,uyy,uyz,uzz,det,&
               psi4pt*gxx(i,j,k),psi4pt*gxy(i,j,k),psi4pt*gxz(i,j,k),&
               psi4pt*gyy(i,j,k),psi4pt*gyz(i,j,k),psi4pt*gzz(i,j,k))
        end if

        call Con2Prim_ptPolytype(whisky_eos_handle, dens(i,j,k),&
             sx(i,j,k),sy(i,j,k), &
             sz(i,j,k),tau(i,j,k),rho(i,j,k),velx(i,j,k),vely(i,j,k), &
             velz(i,j,k),eps(i,j,k),press(i,j,k),w_lorentz(i,j,k), &
             uxx,uxy,uxz,uyy,uyz,uzz,det,x(i,j,k),y(i,j,k), &
             z(i,j,k),r(i,j,k),whisky_rho_min)
        
      end do
    end do
  end do
  
end subroutine Conservative2PrimitivePolytype

/*@@
@routine    Con2Prim_ptPolytype
@date       Sat Jan 26 01:09:39 2002
@author     
@desc 
Given all the appropriate data, recover the primitive variables at
a single point.
@enddesc 
 @calls     
@calledby   
@history 

@endhistory 

@@*/

subroutine Con2Prim_ptPolytype(handle, dens, sx, sy, sz, tau, rho, &
     velx, vely, velz, epsilon, press, w_lorentz, uxx, uxy, uxz, uyy, &
     uyz, uzz, det, x, y, z, r, whisky_rho_min)
  
  implicit none
  
  DECLARE_CCTK_PARAMETERS

  CCTK_REAL dens, sx, sy, sz, tau, rho, velx, vely, velz, epsilon, &
       press, uxx, uxy, uxz, uyy, uyz, uzz, det, w_lorentz, x, &
       y, z, r, whisky_rho_min
  
  CCTK_REAL s2, f, df, ftol, vlowx, vlowy, vlowz 
  CCTK_REAL dens_prim, tau_prim, sx_prim, sy_prim, sz_prim, test
  CCTK_INT count, i, handle
  CCTK_REAL udens, usx, usy, usz, rhoold, rhonew, epsold, epsnew, w2, &
       w2mhalf, enthalpy, denthalpy
  character(len=100) warnline

#ifdef _EOS_BASE_INC_
#undef _EOS_BASE_INC_
#endif
#include "EOS_Base.inc"

!!$I''ve removed the setting to zero for the T3E that was here.
!!$It was a hack to start with, and you should get the same results
!!$in a cleaner fashion using the parameter 
!!$pugh::initialize_memory="zero"
  
!!$  Undensitize the variables 

  udens = dens /sqrt(det)
  usx = sx /sqrt(det)
  usy = sy /sqrt(det)
  usz = sz /sqrt(det)
  s2 = usx*usx*uxx + usy*usy*uyy + usz*usz*uzz + 2.d0*usx*usy*uxy + &
       2.d0*usx*usz*uxz + 2.d0*usy*usz*uyz
  
!!$  Set initial guess for rho:

  rhoold = max(whisky_rho_min,rho)
  
!!$  Calculate w_lorentz etc. 

  enthalpy = 1.d0 + EOS_SpecificIntEnergy(handle, rhoold, press) + &
       EOS_Pressure(handle, rhoold, epsilon) / rhoold
  w_lorentz = sqrt(1.d0 + s2 / ( (udens*enthalpy)**2 ))
  press = EOS_Pressure(handle, rhoold, epsilon)

!!$  Calculate the function

  f = rhoold * w_lorentz - udens

!!$Find the root
  
  count = 0
  rhonew = rhoold
  do while ( ((abs(rhonew - rhoold)/abs(rhonew) .gt. whisky_perc_ptol) .and. &
       (abs(rhonew - rhoold) .gt. whisky_del_ptol))  .or. &
       (count .lt. whisky_countmin))
    count = count + 1
    if (count .gt. whisky_countmax) then
      call CCTK_WARN(1, 'Con2Prim_ptPolytype(press): error: did not converge in ')
      write(warnline,'(a20,g12.7,a10)') '              ',&
           whisky_countmax,' steps'
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'de is ',dens
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sx is ',sx
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sy is ',sy
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'sz is ',sz
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'rho func: ',f
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'deriv of rho func: ',df
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'rhonew: ',rhonew
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'rhoold: ',rhoold
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'press: ',press
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'epsilon: ',epsilon
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'w_lorentz: ',w_lorentz
      call CCTK_WARN(1,warnline)

      write(warnline,'(a20,3g12.7)') 'xyz location: ',x,y,z
      call CCTK_WARN(1,warnline)
      write(warnline,'(a20,g12.7)') 'radius: ',r
      call CCTK_WARN(1,warnline)
      call CCTK_WARN(0, "Did not converge")
    endif

!!$    I have the feeling that this is an error in general. But for the
!!$    2D_Polytrope it gives the right answers.

    denthalpy = EOS_DPressByDrho(handle, rhonew, epsilon) / rhonew

    df = w_lorentz - rhonew * s2 * denthalpy / &
         (w_lorentz*(udens**2)*(enthalpy**3))

    rhoold = rhonew
    rhonew = rhoold - f/df
    
!!$    Recalculate primitive variables and function
       
    enthalpy = 1.d0 + EOS_SpecificIntEnergy(handle, rhonew, press) + &
         EOS_Pressure(handle, rhonew, epsilon) / rhonew
    w_lorentz = sqrt(1.d0 + s2 / ( (udens*enthalpy)**2 ))
    press = EOS_Pressure(handle, rhonew, epsilon)
    tau = sqrt(det) * ((rho * enthalpy) * w_lorentz**2 - press) - dens
    
    f = rhonew * w_lorentz - udens

  enddo
      
!!$  Calculate primitive variables from root
  if (rhonew .le. whisky_rho_min) then
    rhonew = whisky_rho_min
    sx = 0.d0
    sy = 0.d0
    sz = 0.d0
    s2 = 0.d0
    usx = 0.d0
    usy = 0.d0
    usz = 0.d0
!!$    It is a bad thing to have a consistent dens, apparently
  end if

  rho = rhonew
  enthalpy = 1.d0 + EOS_SpecificIntEnergy(handle, rhonew, press) + &
       EOS_Pressure(handle, rhonew, epsilon) / rhonew
  w_lorentz = sqrt(1.d0 + s2 / ( (udens*enthalpy)**2 ))
  press = EOS_Pressure(handle, rhonew, epsilon)
  epsilon = EOS_SpecificIntEnergy(handle, rhonew, press)
  tau = sqrt(det) * ((rho * enthalpy) * w_lorentz**2 - press) - dens
  vlowx = usx / ( (rho + rho*epsilon + press) * w_lorentz**2)
  vlowy = usy / ( (rho + rho*epsilon + press) * w_lorentz**2)
  vlowz = usz / ( (rho + rho*epsilon + press) * w_lorentz**2)
  velx = uxx * vlowx + uxy * vlowy + uxz * vlowz
  vely = uxy * vlowx + uyy * vlowy + uyz * vlowz
  velz = uxz * vlowx + uyz * vlowy + uzz * vlowz
  
  if(epsilon .lt. 0.0d0) then
    call CCTK_WARN(1,'Con2prim_ptPolytype: stopping the code.')
    call CCTK_WARN(1, '   specific internal energy just went below 0! ')
    write(warnline,'(a20,3g16.7)') 'xyz location: ',x,y,z
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,g16.7)') 'radius: ',r
    call CCTK_WARN(1,warnline)
    write(warnline,'(a20,3g16.7)') 'velocities: ',velx,vely,velz
    call CCTK_WARN(1,warnline)
    call CCTK_WARN(0, "Specific internal energy negative")
  endif

end subroutine Con2Prim_ptPolytype

 /*@@
   @routine    Cons2PrimBoundsPolytype
   @date       Tue Mar 12 18:04:40 2002
   @author     
   @desc 
   
   @enddesc 
   @calls     
   @calledby   
   @history 
 
   @endhistory 

@@*/


subroutine Con2PrimBoundsPolytype(CCTK_ARGUMENTS)
  
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  integer :: i, j, k, nx, ny, nz
  CCTK_REAL :: uxxl, uxyl, uxzl, uyyl, uyzl, uzzl,&
       uxxr, uxyr, uxzr, uyyr, uyzr, uzzr
  CCTK_REAL :: gxxl,gxyl,gxzl,gyyl,gyzl,gzzl,avg_detl,&
       gxxr,gxyr,gxzr,gyyr,gyzr,gzzr,avg_detr,psi4r,psi4l
 
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
  
  
  do k = whisky_stencil, nz - whisky_stencil + 1
    do j = whisky_stencil, ny - whisky_stencil + 1
      do i = whisky_stencil, nx - whisky_stencil + 1
        
        gxxl = 0.5d0 * (gxx(i,j,k) + gxx(i-xoffset,j-yoffset,k-zoffset))
        gxyl = 0.5d0 * (gxy(i,j,k) + gxy(i-xoffset,j-yoffset,k-zoffset))
        gxzl = 0.5d0 * (gxz(i,j,k) + gxz(i-xoffset,j-yoffset,k-zoffset))
        gyyl = 0.5d0 * (gyy(i,j,k) + gyy(i-xoffset,j-yoffset,k-zoffset))
        gyzl = 0.5d0 * (gyz(i,j,k) + gyz(i-xoffset,j-yoffset,k-zoffset))
        gzzl = 0.5d0 * (gzz(i,j,k) + gzz(i-xoffset,j-yoffset,k-zoffset))
        gxxr = 0.5d0 * (gxx(i,j,k) + gxx(i+xoffset,j+yoffset,k+zoffset))
        gxyr = 0.5d0 * (gxy(i,j,k) + gxy(i+xoffset,j+yoffset,k+zoffset))
        gxzr = 0.5d0 * (gxz(i,j,k) + gxz(i+xoffset,j+yoffset,k+zoffset))
        gyyr = 0.5d0 * (gyy(i,j,k) + gyy(i+xoffset,j+yoffset,k+zoffset))
        gyzr = 0.5d0 * (gyz(i,j,k) + gyz(i+xoffset,j+yoffset,k+zoffset))
        gzzr = 0.5d0 * (gzz(i,j,k) + gzz(i+xoffset,j+yoffset,k+zoffset))

        if (conformal_state .eq. 0) then
          call SpatialDeterminant(gxxl,gxyl,gxzl,gyyl, gyzl,gzzl,avg_detl)
          call SpatialDeterminant(gxxr,gxyr,gxzr,gyyr, gyzr,gzzr,avg_detr)
          call UpperMetric(uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
               gxxl,gxyl,gxzl,gyyl,gyzl,gzzl)        
          call UpperMetric(uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
               gxxr,gxyr,gxzr,gyyr,gyzr,gzzr)        
        else
          psi4r = (0.5d0*(psi(i,j,k)+psi(i+xoffset,j+yoffset,k+zoffset)))**4
          psi4l = (0.5d0*(psi(i,j,k)+psi(i-xoffset,j-yoffset,k-zoffset)))**4

          call SpatialDeterminant(psi4l*gxxl,psi4l*gxyl,psi4l*gxzl,&
               psi4l*gyyl,psi4l*gyzl,psi4l*gzzl,avg_detl)
          call SpatialDeterminant(psi4r*gxxr,psi4r*gxyr,psi4r*gxzr,&
               psi4r*gyyr,psi4r*gyzr,psi4r*gzzr,avg_detr)
          call UpperMetric(uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
               psi4l*gxxl,psi4l*gxyl,psi4l*gxzl,psi4l*gyyl,psi4l*gyzl,&
               psi4l*gzzl)        
          call UpperMetric(uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
               psi4r*gxxr,psi4r*gxyr,psi4r*gxzr,psi4r*gyyr,psi4r*gyzr,&
               psi4r*gzzr)        
        end if

        call Con2Prim_ptPolytype(whisky_eos_handle, densminus(i,j,k),&
             sxminus(i,j,k),syminus(i,j,k),szminus(i,j,k),&
             tauminus(i,j,k),rhominus(i,j,k),velxminus(i,j,k),&
             velyminus(i,j,k),velzminus(i,j,k),epsminus(i,j,k),&
             pressminus(i,j,k),w_lorentzminus(i,j,k),&
             uxxl,uxyl,uxzl,uyyl,uyzl,uzzl,avg_detl,&
             x(i,j,k)-0.5d0*CCTK_DELTA_SPACE(1),&
             y(i,j,k)-0.5d0*CCTK_DELTA_SPACE(2), &
             z(i,j,k)-0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),whisky_rho_min)
        call Con2Prim_ptPolytype(whisky_eos_handle, densplus(i,j,k),&
             sxplus(i,j,k),syplus(i,j,k),szplus(i,j,k),&
             tauplus(i,j,k),rhoplus(i,j,k),velxplus(i,j,k),&
             velyplus(i,j,k),velzplus(i,j,k),epsplus(i,j,k),&
             pressplus(i,j,k),w_lorentzplus(i,j,k),&
             uxxr,uxyr,uxzr,uyyr,uyzr,uzzr,avg_detr,&
             x(i,j,k)+0.5d0*CCTK_DELTA_SPACE(1),&
             y(i,j,k)+0.5d0*CCTK_DELTA_SPACE(2), &
             z(i,j,k)+0.5d0*CCTK_DELTA_SPACE(3),r(i,j,k),whisky_rho_min)
        
      end do
    end do
  end do
  
end subroutine Con2PrimBoundsPolytype
