 /*@@
   @file      Whisky_Source.F90
   @date      Sat Jan 26 02:03:56 2002
   @author    Ian Hawke
   @desc 
   The geometric source terms for the matter evolution
   @enddesc 
 @@*/


#define DIFF_X(q) 0.5d0 * (q(i+1,j,k) - q(i-1,j,k)) * idx
#define DIFF_Y(q) 0.5d0 * (q(i,j+1,k) - q(i,j-1,k)) * idy
#define DIFF_Z(q) 0.5d0 * (q(i,j,k+1) - q(i,j,k-1)) * idz

! Fourth order f.d.

!!$#define DIFF_X(q) (-q(i+2,j,k) + 8.d0 * q(i+1,j,k) - 8.d0 * q(i-1,j,k) + \
!!$                    q(i-2,j,k)) / 12.d0
!!$#define DIFF_Y(q) (-q(i,j+2,k) + 8.d0 * q(i,j+1,k) - 8.d0 * q(i,j-1,k) + \
!!$                    q(i,j-2,k)) / 12.d0
!!$#define DIFF_Z(q) (-q(i,j,k+2) + 8.d0 * q(i,j,k+1) - 8.d0 * q(i,j,k-1) + \
!!$                    q(i,j,k-2)) / 12.d0

#include "cctk.h"
#include "cctk_Parameters.h"
#include "cctk_Arguments.h"

 /*@@
   @routine    SourceTerms
   @date       Sat Jan 26 02:04:21 2002
   @author     Ian Hawke
   @desc 
   Calculate the geometric source terms and add to the update GFs
   @enddesc 
   @calls     
   @calledby   
   @history 
   Minor alterations of routine from GR3D.
   @endhistory 

@@*/

subroutine SourceTerms(CCTK_ARGUMENTS)
      
  implicit none
  
  DECLARE_CCTK_ARGUMENTS
  DECLARE_CCTK_PARAMETERS
  
  CCTK_INT :: i, j, k, nx, ny, nz
  CCTK_REAL :: one, two, half
  CCTK_REAL :: t00, t0x, t0y, t0z, txx, txy, txz, tyy, tyz, tzz
  CCTK_REAL :: sqrtdet, det, uxx, uxy, uxz, uyy, uyz, uzz, rhoenthalpyW2
  CCTK_REAL :: shiftx, shifty, shiftz, velxshift, velyshift, velzshift 
  CCTK_REAL :: vlowx, vlowy, vlowz
  CCTK_REAL :: dx_betax, dx_betay, dx_betaz, dy_betax, dy_betay,&
       dy_betaz, dz_betax, dz_betay, dz_betaz
  CCTK_REAL :: dx_alp, dy_alp, dz_alp
  CCTK_REAL :: tau_source, sx_source, sy_source, sz_source
  CCTK_REAL :: localgxx,localgxy,localgxz,localgyy,localgyz,localgzz
  CCTK_REAL :: dx_gxx, dx_gxy, dx_gxz, dx_gyy, dx_gyz, dx_gzz
  CCTK_REAL :: dy_gxx, dy_gxy, dy_gxz, dy_gyy, dy_gyz, dy_gzz
  CCTK_REAL :: dz_gxx, dz_gxy, dz_gxz, dz_gyy, dz_gyz, dz_gzz
  CCTK_REAL :: dx, dy, dz, idx, idy, idz
  CCTK_REAL :: psi4pt, dx_psi4, dy_psi4, dz_psi4
  CCTK_REAL :: shiftshiftk, shiftkx, shiftky, shiftkz
  CCTK_REAL :: sumTK
  CCTK_REAL :: halfshiftdgx, halfshiftdgy, halfshiftdgz
  CCTK_REAL :: halfTdgx, halfTdgy, halfTdgz

  one = 1.0d0
  two = 2.0d0
  half = 0.5d0
  nx = cctk_lsh(1)
  ny = cctk_lsh(2)
  nz = cctk_lsh(3)
  dx = CCTK_DELTA_SPACE(1)
  dy = CCTK_DELTA_SPACE(2)
  dz = CCTK_DELTA_SPACE(3)
  idx = 1.d0/dx
  idy = 1.d0/dy
  idz = 1.d0/dz
 
!!$  Initialize the update terms to be zero.
!!$  This will guarantee that no garbage in the boundaries is updated.

  densrhs = 0.d0
  sxrhs = 0.d0
  syrhs = 0.d0
  szrhs = 0.d0
  taurhs = 0.d0
  
  do k=1 + whisky_stencil,nz - whisky_stencil
    do j=1 + whisky_stencil,ny - whisky_stencil
      do i=1 + whisky_stencil,nx - whisky_stencil

!!$        Set the metric terms.

        if (conformal_state .eq. 0) then

          localgxx = gxx(i,j,k)
          localgxy = gxy(i,j,k)
          localgxz = gxz(i,j,k)
          localgyy = gyy(i,j,k)
          localgyz = gyz(i,j,k)
          localgzz = gzz(i,j,k)

        else

          psi4pt = psi(i,j,k)**4
          
          localgxx = psi4pt*gxx(i,j,k)
          localgxy = psi4pt*gxy(i,j,k)
          localgxz = psi4pt*gxz(i,j,k)
          localgyy = psi4pt*gyy(i,j,k)
          localgyz = psi4pt*gyz(i,j,k)
          localgzz = psi4pt*gzz(i,j,k)
          
        end if


        call SpatialDeterminant(localgxx, localgxy, localgxz,&
             localgyy, localgyz, localgzz, det)
        sqrtdet = sqrt(det)
        call UpperMetric(uxx, uxy, uxz, uyy, uyz, uzz, det, localgxx,&
             localgxy, localgxz, localgyy, localgyz, localgzz)
        
!!$        All the matter variables (except velocity) always appear
!!$        together in this form

        rhoenthalpyW2 = (rho(i,j,k)*(one + eps(i,j,k)) + press(i,j,k))*&
             w_lorentz(i,j,k)**2
        
        if (shift_state .ne. 0) then 

          shiftx = betax(i,j,k)
          shifty = betay(i,j,k)
          shiftz = betaz(i,j,k)

          dx_betax = DIFF_X(betax)
          dx_betay = DIFF_X(betay)
          dx_betaz = DIFF_X(betaz)

          dy_betax = DIFF_Y(betax)
          dy_betay = DIFF_Y(betay)
          dy_betaz = DIFF_Y(betaz)

          dz_betax = DIFF_Z(betax)
          dz_betay = DIFF_Z(betay)
          dz_betaz = DIFF_Z(betaz)

        else

          shiftx = 0.0d0
          shifty = 0.0d0
          shiftz = 0.0d0

          dx_betax = 0.0d0
          dx_betay = 0.0d0
          dx_betaz = 0.0d0
          
          dy_betax = 0.0d0
          dy_betay = 0.0d0
          dy_betaz = 0.0d0
          
          dz_betax = 0.0d0
          dz_betay = 0.0d0
          dz_betaz = 0.0d0
          
        endif

        velxshift = velx(i,j,k) - shiftx/alp(i,j,k)
        velyshift = vely(i,j,k) - shifty/alp(i,j,k)
        velzshift = velz(i,j,k) - shiftz/alp(i,j,k)
        vlowx = velx(i,j,k)*localgxx + vely(i,j,k)*localgxy +&
             velz(i,j,k)*localgxz
        vlowy = velx(i,j,k)*localgxy + vely(i,j,k)*localgyy +&
             velz(i,j,k)*localgyz
        vlowz = velx(i,j,k)*localgxz + vely(i,j,k)*localgyz +&
             velz(i,j,k)*localgzz

!!$        For a change, these are T^{ij}

        t00 = (rhoenthalpyW2 - press(i,j,k))/(alp(i,j,k)**2)
        t0x = rhoenthalpyW2*velxshift/alp(i,j,k) +&
             press(i,j,k)*shiftx/(alp(i,j,k)**2)
        t0y = rhoenthalpyW2*velyshift/alp(i,j,k) +&
             press(i,j,k)*shifty/(alp(i,j,k)**2)
        t0z = rhoenthalpyW2*velzshift/alp(i,j,k) +&
             press(i,j,k)*shiftz/(alp(i,j,k)**2)
        txx = rhoenthalpyW2*velxshift*velxshift +&
             press(i,j,k)*(uxx - shiftx*shiftx/(alp(i,j,k)**2))
        txy = rhoenthalpyW2*velxshift*velyshift +&
             press(i,j,k)*(uxy - shiftx*shifty/(alp(i,j,k)**2))
        txz = rhoenthalpyW2*velxshift*velzshift +&
             press(i,j,k)*(uxz - shiftx*shiftz/(alp(i,j,k)**2))
        tyy = rhoenthalpyW2*velyshift*velyshift +&
             press(i,j,k)*(uyy - shifty*shifty/(alp(i,j,k)**2))
        tyz = rhoenthalpyW2*velyshift*velzshift +&
             press(i,j,k)*(uyz - shifty*shiftz/(alp(i,j,k)**2))
        tzz = rhoenthalpyW2*velzshift*velzshift +&
             press(i,j,k)*(uzz - shiftz*shiftz/(alp(i,j,k)**2))

!!$        Derivatives of the lapse, metric and shift

        dx_alp = DIFF_X(alp)
        dy_alp = DIFF_Y(alp)
        dz_alp = DIFF_Z(alp)

        if (conformal_state .eq. 0) then

          dx_gxx = DIFF_X(gxx)
          dx_gxy = DIFF_X(gxy)
          dx_gxz = DIFF_X(gxz)
          dx_gyy = DIFF_X(gyy)
          dx_gyz = DIFF_X(gyz)
          dx_gzz = DIFF_X(gzz)
          dy_gxx = DIFF_Y(gxx)
          dy_gxy = DIFF_Y(gxy)
          dy_gxz = DIFF_Y(gxz)
          dy_gyy = DIFF_Y(gyy)
          dy_gyz = DIFF_Y(gyz)
          dy_gzz = DIFF_Y(gzz)
          dz_gxx = DIFF_Z(gxx)
          dz_gxy = DIFF_Z(gxy)
          dz_gxz = DIFF_Z(gxz)
          dz_gyy = DIFF_Z(gyy)
          dz_gyz = DIFF_Z(gyz)
          dz_gzz = DIFF_Z(gzz)

        else if (conformal_state > 1) then

!!$          Using conformal factor, and the derivatives are
!!$          already calculated.

          psi4pt = psi(i,j,k)**4
          dx_psi4 = 4.d0*psi4pt*psix(i,j,k)
          dy_psi4 = 4.d0*psi4pt*psiy(i,j,k)
          dz_psi4 = 4.d0*psi4pt*psiz(i,j,k)

          dx_gxx = DIFF_X(gxx)*psi4pt + dx_psi4*gxx(i,j,k)
          dx_gxy = DIFF_X(gxy)*psi4pt + dx_psi4*gxy(i,j,k)
          dx_gxz = DIFF_X(gxz)*psi4pt + dx_psi4*gxz(i,j,k)
          dx_gyy = DIFF_X(gyy)*psi4pt + dx_psi4*gyy(i,j,k)
          dx_gyz = DIFF_X(gyz)*psi4pt + dx_psi4*gyz(i,j,k)
          dx_gzz = DIFF_X(gzz)*psi4pt + dx_psi4*gzz(i,j,k)
          dy_gxx = DIFF_Y(gxx)*psi4pt + dy_psi4*gxx(i,j,k)
          dy_gxy = DIFF_Y(gxy)*psi4pt + dy_psi4*gxy(i,j,k)
          dy_gxz = DIFF_Y(gxz)*psi4pt + dy_psi4*gxz(i,j,k)
          dy_gyy = DIFF_Y(gyy)*psi4pt + dy_psi4*gyy(i,j,k)
          dy_gyz = DIFF_Y(gyz)*psi4pt + dy_psi4*gyz(i,j,k)
          dy_gzz = DIFF_Y(gzz)*psi4pt + dy_psi4*gzz(i,j,k)
          dz_gxx = DIFF_Z(gxx)*psi4pt + dz_psi4*gxx(i,j,k)
          dz_gxy = DIFF_Z(gxy)*psi4pt + dz_psi4*gxy(i,j,k)
          dz_gxz = DIFF_Z(gxz)*psi4pt + dz_psi4*gxz(i,j,k)
          dz_gyy = DIFF_Z(gyy)*psi4pt + dz_psi4*gyy(i,j,k)
          dz_gyz = DIFF_Z(gyz)*psi4pt + dz_psi4*gyz(i,j,k)
          dz_gzz = DIFF_Z(gzz)*psi4pt + dz_psi4*gzz(i,j,k)
        else
!!$          Have to finite difference the conformal factor
          call CCTK_WARN(0,"At this point in the source terms we should be differencing the conformal factor. Ian has been too lazy to put this bit of code in yet.")
        end if

!!$        Contract the shift with the extrinsic curvature

        shiftshiftk = shiftx*shiftx*kxx(i,j,k) + &
                      shifty*shifty*kyy(i,j,k) + &
                      shiftz*shiftz*kzz(i,j,k) + &
             two*(shiftx*shifty*kxy(i,j,k) + &
                  shiftx*shiftz*kxz(i,j,k) + &
                  shifty*shiftz*kyz(i,j,k))

        shiftkx = shiftx*kxx(i,j,k) + shifty*kxy(i,j,k) + shiftz*kxz(i,j,k)
        shiftky = shiftx*kxy(i,j,k) + shifty*kyy(i,j,k) + shiftz*kyz(i,j,k)
        shiftkz = shiftx*kxz(i,j,k) + shifty*kyz(i,j,k) + shiftz*kzz(i,j,k)

!!$        Contract the matter terms with the extrinsic curvature

        sumTK = txx*kxx(i,j,k) + tyy*kyy(i,j,k) + tzz*kzz(i,j,k) &
             + two*(txy*kxy(i,j,k) + txz*kxz(i,j,k) + tyz*kyz(i,j,k))

!!$        Update term for tau
        
        tau_source = t00* &
             (shiftshiftk - (shiftx*dx_alp + shifty*dy_alp + shiftz*dz_alp) )&
             + t0x*(-dx_alp + two*shiftkx) &
             + t0y*(-dy_alp + two*shiftky) &
             + t0z*(-dz_alp + two*shiftkz) &
             + sumTK

!!$        The following looks very little like the terms in the
!!$        standard papers. Take a look in the ThornGuide to see why
!!$        it is really the same thing.

!!$        Contract the shift with derivatives of the metric

        halfshiftdgx = half*(shiftx*shiftx*dx_gxx + &
             shifty*shifty*dx_gyy + shiftz*shiftz*dx_gzz) + &
             shiftx*shifty*dx_gxy + shiftx*shiftz*dx_gxz + &
             shifty*shiftz*dx_gyz
        halfshiftdgy = half*(shiftx*shiftx*dy_gxx + &
             shifty*shifty*dy_gyy + shiftz*shiftz*dy_gzz) + &
             shiftx*shifty*dy_gxy + shiftx*shiftz*dy_gxz + &
             shifty*shiftz*dy_gyz
        halfshiftdgz = half*(shiftx*shiftx*dz_gxx + &
             shifty*shifty*dz_gyy + shiftz*shiftz*dz_gzz) + &
             shiftx*shifty*dz_gxy + shiftx*shiftz*dz_gxz + &
             shifty*shiftz*dz_gyz

!!$        Contract the matter with derivatives of the metric

        halfTdgx = half*(txx*dx_gxx + tyy*dx_gyy + tzz*dx_gzz) +&
             txy*dx_gxy + txz*dx_gxz + tyz*dx_gyz
        halfTdgy = half*(txx*dy_gxx + tyy*dy_gyy + tzz*dy_gzz) +&
             txy*dy_gxy + txz*dy_gxz + tyz*dy_gyz
        halfTdgz = half*(txx*dz_gxx + tyy*dz_gyy + tzz*dz_gzz) +&
             txy*dz_gxy + txz*dz_gxz + tyz*dz_gyz

        sx_source = t00*&
             (halfshiftdgx - alp(i,j,k)*dx_alp) +&
             t0x*(shiftx*dx_gxx + shifty*dx_gxy + shiftz*dx_gxz) +&
             t0y*(shiftx*dx_gxy + shifty*dx_gyy + shiftz*dx_gyz) +&
             t0z*(shiftx*dx_gxz + shifty*dx_gyz + shiftz*dx_gzz) +&
             halfTdgx + rhoenthalpyW2*&
             (vlowx*dx_betax + vlowy*dx_betay + vlowz*dx_betaz)/&
             alp(i,j,k)
        sy_source = t00*&
             (halfshiftdgy - alp(i,j,k)*dy_alp) +&
             t0x*(shiftx*dy_gxx + shifty*dy_gxy + shiftz*dy_gxz) +&
             t0y*(shiftx*dy_gxy + shifty*dy_gyy + shiftz*dy_gyz) +&
             t0z*(shiftx*dy_gxz + shifty*dy_gyz + shiftz*dy_gzz) +&
             halfTdgy + rhoenthalpyW2*&
             (vlowx*dy_betax + vlowy*dy_betay + vlowz*dy_betaz)/&
             alp(i,j,k)
        sz_source = t00*&
             (halfshiftdgz - alp(i,j,k)*dz_alp) +&
             t0x*(shiftx*dz_gxx + shifty*dz_gxy + shiftz*dz_gxz) +&
             t0y*(shiftx*dz_gxy + shifty*dz_gyy + shiftz*dz_gyz) +&
             t0z*(shiftx*dz_gxz + shifty*dz_gyz + shiftz*dz_gzz) +&
             halfTdgz + rhoenthalpyW2*&
             (vlowx*dz_betax + vlowy*dz_betay + vlowz*dz_betaz)/&
             alp(i,j,k)

        densrhs(i,j,k) = 0.d0
        sxrhs(i,j,k)  = alp(i,j,k)*sqrtdet*sx_source
        syrhs(i,j,k)  = alp(i,j,k)*sqrtdet*sy_source
        szrhs(i,j,k)  = alp(i,j,k)*sqrtdet*sz_source
        taurhs(i,j,k) = alp(i,j,k)*sqrtdet*tau_source
        
      enddo
    enddo
  enddo

end subroutine SourceTerms



